#define _CFFI_

/* We try to define Py_LIMITED_API before including Python.h.

   Mess: we can only define it if Py_DEBUG, Py_TRACE_REFS and
   Py_REF_DEBUG are not defined.  This is a best-effort approximation:
   we can learn about Py_DEBUG from pyconfig.h, but it is unclear if
   the same works for the other two macros.  Py_DEBUG implies them,
   but not the other way around.

   The implementation is messy (issue #350): on Windows, with _MSC_VER,
   we have to define Py_LIMITED_API even before including pyconfig.h.
   In that case, we guess what pyconfig.h will do to the macros above,
   and check our guess after the #include.

   Note that on Windows, with CPython 3.x, you need >= 3.5 and virtualenv
   version >= 16.0.0.  With older versions of either, you don't get a
   copy of PYTHON3.DLL in the virtualenv.  We can't check the version of
   CPython *before* we even include pyconfig.h.  ffi.set_source() puts
   a ``#define _CFFI_NO_LIMITED_API'' at the start of this file if it is
   running on Windows < 3.5, as an attempt at fixing it, but that's
   arguably wrong because it may not be the target version of Python.
   Still better than nothing I guess.  As another workaround, you can
   remove the definition of Py_LIMITED_API here.

   See also 'py_limited_api' in cffi/setuptools_ext.py.
*/
#if !defined(_CFFI_USE_EMBEDDING) && !defined(Py_LIMITED_API)
#  ifdef _MSC_VER
#    if !defined(_DEBUG) && !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG) && !defined(_CFFI_NO_LIMITED_API)
#      define Py_LIMITED_API
#    endif
#    include <pyconfig.h>
     /* sanity-check: Py_LIMITED_API will cause crashes if any of these
        are also defined.  Normally, the Python file PC/pyconfig.h does not
        cause any of these to be defined, with the exception that _DEBUG
        causes Py_DEBUG.  Double-check that. */
#    ifdef Py_LIMITED_API
#      if defined(Py_DEBUG)
#        error "pyconfig.h unexpectedly defines Py_DEBUG, but Py_LIMITED_API is set"
#      endif
#      if defined(Py_TRACE_REFS)
#        error "pyconfig.h unexpectedly defines Py_TRACE_REFS, but Py_LIMITED_API is set"
#      endif
#      if defined(Py_REF_DEBUG)
#        error "pyconfig.h unexpectedly defines Py_REF_DEBUG, but Py_LIMITED_API is set"
#      endif
#    endif
#  else
#    include <pyconfig.h>
#    if !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG) && !defined(_CFFI_NO_LIMITED_API)
#      define Py_LIMITED_API
#    endif
#  endif
#endif

#include <Python.h>
#ifdef __cplusplus
extern "C" {
#endif
#include <stddef.h>

/* This part is from file 'cffi/parse_c_type.h'.  It is copied at the
   beginning of C sources generated by CFFI's ffi.set_source(). */

typedef void *_cffi_opcode_t;

#define _CFFI_OP(opcode, arg)   (_cffi_opcode_t)(opcode | (((uintptr_t)(arg)) << 8))
#define _CFFI_GETOP(cffi_opcode)    ((unsigned char)(uintptr_t)cffi_opcode)
#define _CFFI_GETARG(cffi_opcode)   (((intptr_t)cffi_opcode) >> 8)

#define _CFFI_OP_PRIMITIVE       1
#define _CFFI_OP_POINTER         3
#define _CFFI_OP_ARRAY           5
#define _CFFI_OP_OPEN_ARRAY      7
#define _CFFI_OP_STRUCT_UNION    9
#define _CFFI_OP_ENUM           11
#define _CFFI_OP_FUNCTION       13
#define _CFFI_OP_FUNCTION_END   15
#define _CFFI_OP_NOOP           17
#define _CFFI_OP_BITFIELD       19
#define _CFFI_OP_TYPENAME       21
#define _CFFI_OP_CPYTHON_BLTN_V 23   // varargs
#define _CFFI_OP_CPYTHON_BLTN_N 25   // noargs
#define _CFFI_OP_CPYTHON_BLTN_O 27   // O  (i.e. a single arg)
#define _CFFI_OP_CONSTANT       29
#define _CFFI_OP_CONSTANT_INT   31
#define _CFFI_OP_GLOBAL_VAR     33
#define _CFFI_OP_DLOPEN_FUNC    35
#define _CFFI_OP_DLOPEN_CONST   37
#define _CFFI_OP_GLOBAL_VAR_F   39
#define _CFFI_OP_EXTERN_PYTHON  41

#define _CFFI_PRIM_VOID          0
#define _CFFI_PRIM_BOOL          1
#define _CFFI_PRIM_CHAR          2
#define _CFFI_PRIM_SCHAR         3
#define _CFFI_PRIM_UCHAR         4
#define _CFFI_PRIM_SHORT         5
#define _CFFI_PRIM_USHORT        6
#define _CFFI_PRIM_INT           7
#define _CFFI_PRIM_UINT          8
#define _CFFI_PRIM_LONG          9
#define _CFFI_PRIM_ULONG        10
#define _CFFI_PRIM_LONGLONG     11
#define _CFFI_PRIM_ULONGLONG    12
#define _CFFI_PRIM_FLOAT        13
#define _CFFI_PRIM_DOUBLE       14
#define _CFFI_PRIM_LONGDOUBLE   15

#define _CFFI_PRIM_WCHAR        16
#define _CFFI_PRIM_INT8         17
#define _CFFI_PRIM_UINT8        18
#define _CFFI_PRIM_INT16        19
#define _CFFI_PRIM_UINT16       20
#define _CFFI_PRIM_INT32        21
#define _CFFI_PRIM_UINT32       22
#define _CFFI_PRIM_INT64        23
#define _CFFI_PRIM_UINT64       24
#define _CFFI_PRIM_INTPTR       25
#define _CFFI_PRIM_UINTPTR      26
#define _CFFI_PRIM_PTRDIFF      27
#define _CFFI_PRIM_SIZE         28
#define _CFFI_PRIM_SSIZE        29
#define _CFFI_PRIM_INT_LEAST8   30
#define _CFFI_PRIM_UINT_LEAST8  31
#define _CFFI_PRIM_INT_LEAST16  32
#define _CFFI_PRIM_UINT_LEAST16 33
#define _CFFI_PRIM_INT_LEAST32  34
#define _CFFI_PRIM_UINT_LEAST32 35
#define _CFFI_PRIM_INT_LEAST64  36
#define _CFFI_PRIM_UINT_LEAST64 37
#define _CFFI_PRIM_INT_FAST8    38
#define _CFFI_PRIM_UINT_FAST8   39
#define _CFFI_PRIM_INT_FAST16   40
#define _CFFI_PRIM_UINT_FAST16  41
#define _CFFI_PRIM_INT_FAST32   42
#define _CFFI_PRIM_UINT_FAST32  43
#define _CFFI_PRIM_INT_FAST64   44
#define _CFFI_PRIM_UINT_FAST64  45
#define _CFFI_PRIM_INTMAX       46
#define _CFFI_PRIM_UINTMAX      47
#define _CFFI_PRIM_FLOATCOMPLEX 48
#define _CFFI_PRIM_DOUBLECOMPLEX 49
#define _CFFI_PRIM_CHAR16       50
#define _CFFI_PRIM_CHAR32       51

#define _CFFI__NUM_PRIM         52
#define _CFFI__UNKNOWN_PRIM           (-1)
#define _CFFI__UNKNOWN_FLOAT_PRIM     (-2)
#define _CFFI__UNKNOWN_LONG_DOUBLE    (-3)

#define _CFFI__IO_FILE_STRUCT         (-1)


struct _cffi_global_s {
    const char *name;
    void *address;
    _cffi_opcode_t type_op;
    void *size_or_direct_fn;  // OP_GLOBAL_VAR: size, or 0 if unknown
                              // OP_CPYTHON_BLTN_*: addr of direct function
};

struct _cffi_getconst_s {
    unsigned long long value;
    const struct _cffi_type_context_s *ctx;
    int gindex;
};

struct _cffi_struct_union_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_STRUCT_UNION
    int flags;               // _CFFI_F_* flags below
    size_t size;
    int alignment;
    int first_field_index;   // -> _cffi_fields array
    int num_fields;
};
#define _CFFI_F_UNION         0x01   // is a union, not a struct
#define _CFFI_F_CHECK_FIELDS  0x02   // complain if fields are not in the
                                     // "standard layout" or if some are missing
#define _CFFI_F_PACKED        0x04   // for CHECK_FIELDS, assume a packed struct
#define _CFFI_F_EXTERNAL      0x08   // in some other ffi.include()
#define _CFFI_F_OPAQUE        0x10   // opaque

struct _cffi_field_s {
    const char *name;
    size_t field_offset;
    size_t field_size;
    _cffi_opcode_t field_type_op;
};

struct _cffi_enum_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_ENUM
    int type_prim;           // _CFFI_PRIM_xxx
    const char *enumerators; // comma-delimited string
};

struct _cffi_typename_s {
    const char *name;
    int type_index;   /* if opaque, points to a possibly artificial
                         OP_STRUCT which is itself opaque */
};

struct _cffi_type_context_s {
    _cffi_opcode_t *types;
    const struct _cffi_global_s *globals;
    const struct _cffi_field_s *fields;
    const struct _cffi_struct_union_s *struct_unions;
    const struct _cffi_enum_s *enums;
    const struct _cffi_typename_s *typenames;
    int num_globals;
    int num_struct_unions;
    int num_enums;
    int num_typenames;
    const char *const *includes;
    int num_types;
    int flags;      /* future extension */
};

struct _cffi_parse_info_s {
    const struct _cffi_type_context_s *ctx;
    _cffi_opcode_t *output;
    unsigned int output_size;
    size_t error_location;
    const char *error_message;
};

struct _cffi_externpy_s {
    const char *name;
    size_t size_of_result;
    void *reserved1, *reserved2;
};

#ifdef _CFFI_INTERNAL
static int parse_c_type(struct _cffi_parse_info_s *info, const char *input);
static int search_in_globals(const struct _cffi_type_context_s *ctx,
                             const char *search, size_t search_len);
static int search_in_struct_unions(const struct _cffi_type_context_s *ctx,
                                   const char *search, size_t search_len);
#endif

/* this block of #ifs should be kept exactly identical between
   c/_cffi_backend.c, cffi/vengine_cpy.py, cffi/vengine_gen.py
   and cffi/_cffi_include.h */
#if defined(_MSC_VER)
# include <malloc.h>   /* for alloca() */
# if _MSC_VER < 1600   /* MSVC < 2010 */
   typedef __int8 int8_t;
   typedef __int16 int16_t;
   typedef __int32 int32_t;
   typedef __int64 int64_t;
   typedef unsigned __int8 uint8_t;
   typedef unsigned __int16 uint16_t;
   typedef unsigned __int32 uint32_t;
   typedef unsigned __int64 uint64_t;
   typedef __int8 int_least8_t;
   typedef __int16 int_least16_t;
   typedef __int32 int_least32_t;
   typedef __int64 int_least64_t;
   typedef unsigned __int8 uint_least8_t;
   typedef unsigned __int16 uint_least16_t;
   typedef unsigned __int32 uint_least32_t;
   typedef unsigned __int64 uint_least64_t;
   typedef __int8 int_fast8_t;
   typedef __int16 int_fast16_t;
   typedef __int32 int_fast32_t;
   typedef __int64 int_fast64_t;
   typedef unsigned __int8 uint_fast8_t;
   typedef unsigned __int16 uint_fast16_t;
   typedef unsigned __int32 uint_fast32_t;
   typedef unsigned __int64 uint_fast64_t;
   typedef __int64 intmax_t;
   typedef unsigned __int64 uintmax_t;
# else
#  include <stdint.h>
# endif
# if _MSC_VER < 1800   /* MSVC < 2013 */
#  ifndef __cplusplus
    typedef unsigned char _Bool;
#  endif
# endif
#else
# include <stdint.h>
# if (defined (__SVR4) && defined (__sun)) || defined(_AIX) || defined(__hpux)
#  include <alloca.h>
# endif
#endif

#ifdef __GNUC__
# define _CFFI_UNUSED_FN  __attribute__((unused))
#else
# define _CFFI_UNUSED_FN  /* nothing */
#endif

#ifdef __cplusplus
# ifndef _Bool
   typedef bool _Bool;   /* semi-hackish: C++ has no _Bool; bool is builtin */
# endif
#endif

/**********  CPython-specific section  **********/
#ifndef PYPY_VERSION


#if PY_MAJOR_VERSION >= 3
# define PyInt_FromLong PyLong_FromLong
#endif

#define _cffi_from_c_double PyFloat_FromDouble
#define _cffi_from_c_float PyFloat_FromDouble
#define _cffi_from_c_long PyInt_FromLong
#define _cffi_from_c_ulong PyLong_FromUnsignedLong
#define _cffi_from_c_longlong PyLong_FromLongLong
#define _cffi_from_c_ulonglong PyLong_FromUnsignedLongLong
#define _cffi_from_c__Bool PyBool_FromLong

#define _cffi_to_c_double PyFloat_AsDouble
#define _cffi_to_c_float PyFloat_AsDouble

#define _cffi_from_c_int(x, type)                                        \
    (((type)-1) > 0 ? /* unsigned */                                     \
        (sizeof(type) < sizeof(long) ?                                   \
            PyInt_FromLong((long)x) :                                    \
         sizeof(type) == sizeof(long) ?                                  \
            PyLong_FromUnsignedLong((unsigned long)x) :                  \
            PyLong_FromUnsignedLongLong((unsigned long long)x)) :        \
        (sizeof(type) <= sizeof(long) ?                                  \
            PyInt_FromLong((long)x) :                                    \
            PyLong_FromLongLong((long long)x)))

#define _cffi_to_c_int(o, type)                                          \
    ((type)(                                                             \
     sizeof(type) == 1 ? (((type)-1) > 0 ? (type)_cffi_to_c_u8(o)        \
                                         : (type)_cffi_to_c_i8(o)) :     \
     sizeof(type) == 2 ? (((type)-1) > 0 ? (type)_cffi_to_c_u16(o)       \
                                         : (type)_cffi_to_c_i16(o)) :    \
     sizeof(type) == 4 ? (((type)-1) > 0 ? (type)_cffi_to_c_u32(o)       \
                                         : (type)_cffi_to_c_i32(o)) :    \
     sizeof(type) == 8 ? (((type)-1) > 0 ? (type)_cffi_to_c_u64(o)       \
                                         : (type)_cffi_to_c_i64(o)) :    \
     (Py_FatalError("unsupported size for type " #type), (type)0)))

#define _cffi_to_c_i8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[1])
#define _cffi_to_c_u8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[2])
#define _cffi_to_c_i16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[3])
#define _cffi_to_c_u16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[4])
#define _cffi_to_c_i32                                                   \
                 ((int(*)(PyObject *))_cffi_exports[5])
#define _cffi_to_c_u32                                                   \
                 ((unsigned int(*)(PyObject *))_cffi_exports[6])
#define _cffi_to_c_i64                                                   \
                 ((long long(*)(PyObject *))_cffi_exports[7])
#define _cffi_to_c_u64                                                   \
                 ((unsigned long long(*)(PyObject *))_cffi_exports[8])
#define _cffi_to_c_char                                                  \
                 ((int(*)(PyObject *))_cffi_exports[9])
#define _cffi_from_c_pointer                                             \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[10])
#define _cffi_to_c_pointer                                               \
    ((char *(*)(PyObject *, struct _cffi_ctypedescr *))_cffi_exports[11])
#define _cffi_get_struct_layout                                          \
    not used any more
#define _cffi_restore_errno                                              \
    ((void(*)(void))_cffi_exports[13])
#define _cffi_save_errno                                                 \
    ((void(*)(void))_cffi_exports[14])
#define _cffi_from_c_char                                                \
    ((PyObject *(*)(char))_cffi_exports[15])
#define _cffi_from_c_deref                                               \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[16])
#define _cffi_to_c                                                       \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[17])
#define _cffi_from_c_struct                                              \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[18])
#define _cffi_to_c_wchar_t                                               \
    ((_cffi_wchar_t(*)(PyObject *))_cffi_exports[19])
#define _cffi_from_c_wchar_t                                             \
    ((PyObject *(*)(_cffi_wchar_t))_cffi_exports[20])
#define _cffi_to_c_long_double                                           \
    ((long double(*)(PyObject *))_cffi_exports[21])
#define _cffi_to_c__Bool                                                 \
    ((_Bool(*)(PyObject *))_cffi_exports[22])
#define _cffi_prepare_pointer_call_argument                              \
    ((Py_ssize_t(*)(struct _cffi_ctypedescr *,                           \
                    PyObject *, char **))_cffi_exports[23])
#define _cffi_convert_array_from_object                                  \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[24])
#define _CFFI_CPIDX  25
#define _cffi_call_python                                                \
    ((void(*)(struct _cffi_externpy_s *, char *))_cffi_exports[_CFFI_CPIDX])
#define _cffi_to_c_wchar3216_t                                           \
    ((int(*)(PyObject *))_cffi_exports[26])
#define _cffi_from_c_wchar3216_t                                         \
    ((PyObject *(*)(int))_cffi_exports[27])
#define _CFFI_NUM_EXPORTS 28

struct _cffi_ctypedescr;

static void *_cffi_exports[_CFFI_NUM_EXPORTS];

#define _cffi_type(index)   (                           \
    assert((((uintptr_t)_cffi_types[index]) & 1) == 0), \
    (struct _cffi_ctypedescr *)_cffi_types[index])

static PyObject *_cffi_init(const char *module_name, Py_ssize_t version,
                            const struct _cffi_type_context_s *ctx)
{
    PyObject *module, *o_arg, *new_module;
    void *raw[] = {
        (void *)module_name,
        (void *)version,
        (void *)_cffi_exports,
        (void *)ctx,
    };

    module = PyImport_ImportModule("_cffi_backend");
    if (module == NULL)
        goto failure;

    o_arg = PyLong_FromVoidPtr((void *)raw);
    if (o_arg == NULL)
        goto failure;

    new_module = PyObject_CallMethod(
        module, (char *)"_init_cffi_1_0_external_module", (char *)"O", o_arg);

    Py_DECREF(o_arg);
    Py_DECREF(module);
    return new_module;

  failure:
    Py_XDECREF(module);
    return NULL;
}


#ifdef HAVE_WCHAR_H
typedef wchar_t _cffi_wchar_t;
#else
typedef uint16_t _cffi_wchar_t;   /* same random pick as _cffi_backend.c */
#endif

_CFFI_UNUSED_FN static uint16_t _cffi_to_c_char16_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return (uint16_t)_cffi_to_c_wchar_t(o);
    else
        return (uint16_t)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char16_t(uint16_t x)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

_CFFI_UNUSED_FN static int _cffi_to_c_char32_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return (int)_cffi_to_c_wchar_t(o);
    else
        return (int)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char32_t(unsigned int x)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

union _cffi_union_alignment_u {
    unsigned char m_char;
    unsigned short m_short;
    unsigned int m_int;
    unsigned long m_long;
    unsigned long long m_longlong;
    float m_float;
    double m_double;
    long double m_longdouble;
};

struct _cffi_freeme_s {
    struct _cffi_freeme_s *next;
    union _cffi_union_alignment_u alignment;
};

_CFFI_UNUSED_FN static int
_cffi_convert_array_argument(struct _cffi_ctypedescr *ctptr, PyObject *arg,
                             char **output_data, Py_ssize_t datasize,
                             struct _cffi_freeme_s **freeme)
{
    char *p;
    if (datasize < 0)
        return -1;

    p = *output_data;
    if (p == NULL) {
        struct _cffi_freeme_s *fp = (struct _cffi_freeme_s *)PyObject_Malloc(
            offsetof(struct _cffi_freeme_s, alignment) + (size_t)datasize);
        if (fp == NULL)
            return -1;
        fp->next = *freeme;
        *freeme = fp;
        p = *output_data = (char *)&fp->alignment;
    }
    memset((void *)p, 0, (size_t)datasize);
    return _cffi_convert_array_from_object(p, ctptr, arg);
}

_CFFI_UNUSED_FN static void
_cffi_free_array_arguments(struct _cffi_freeme_s *freeme)
{
    do {
        void *p = (void *)freeme;
        freeme = freeme->next;
        PyObject_Free(p);
    } while (freeme != NULL);
}

/**********  end CPython-specific section  **********/
#else
_CFFI_UNUSED_FN
static void (*_cffi_call_python_org)(struct _cffi_externpy_s *, char *);
# define _cffi_call_python  _cffi_call_python_org
#endif


#define _cffi_array_len(array)   (sizeof(array) / sizeof((array)[0]))

#define _cffi_prim_int(size, sign)                                      \
    ((size) == 1 ? ((sign) ? _CFFI_PRIM_INT8  : _CFFI_PRIM_UINT8)  :    \
     (size) == 2 ? ((sign) ? _CFFI_PRIM_INT16 : _CFFI_PRIM_UINT16) :    \
     (size) == 4 ? ((sign) ? _CFFI_PRIM_INT32 : _CFFI_PRIM_UINT32) :    \
     (size) == 8 ? ((sign) ? _CFFI_PRIM_INT64 : _CFFI_PRIM_UINT64) :    \
     _CFFI__UNKNOWN_PRIM)

#define _cffi_prim_float(size)                                          \
    ((size) == sizeof(float) ? _CFFI_PRIM_FLOAT :                       \
     (size) == sizeof(double) ? _CFFI_PRIM_DOUBLE :                     \
     (size) == sizeof(long double) ? _CFFI__UNKNOWN_LONG_DOUBLE :       \
     _CFFI__UNKNOWN_FLOAT_PRIM)

#define _cffi_check_int(got, got_nonpos, expected)      \
    ((got_nonpos) == (expected <= 0) &&                 \
     (got) == (unsigned long long)expected)

#ifdef MS_WIN32
# define _cffi_stdcall  __stdcall
#else
# define _cffi_stdcall  /* nothing */
#endif

#ifdef __cplusplus
}
#endif

/************************************************************/

void f_ini_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = -p[14]*x[0] + 314.15926535897933*x[1] - 314.15926535897933*y[72];
out[1] = (-p[11]*(p[13]*x[4] - 0.33333333333333331*y[54]*y[17] - 1.0*y[54]*(-0.16666666666666657*y[19] + 0.28867513459481287*y[18]) - 1.0*y[54]*(-0.16666666666666677*y[21] - 0.28867513459481275*y[20]) - 0.33333333333333331*y[48]*y[16] - 1.0*y[48]*(-0.28867513459481287*y[19] - 0.16666666666666657*y[18]) - 1.0*y[48]*(0.28867513459481275*y[21] - 0.16666666666666677*y[20]) - 0.33333333333333326*y[55]*y[19] + 0.16666666666666652*y[55]*y[21] - 0.28867513459481281*y[55]*y[20] - 0.33333333333333326*y[49]*y[18] + 0.28867513459481281*y[49]*y[21] + 0.16666666666666652*y[49]*y[20] + 0.16666666666666652*y[56]*y[19] + 0.28867513459481281*y[56]*y[18] - 0.33333333333333326*y[56]*y[21] - 0.28867513459481281*y[50]*y[19] + 0.16666666666666652*y[50]*y[18] - 0.33333333333333326*y[50]*y[20] + u[20] + y[17]*(0.16666666666666657*y[55] - 0.28867513459481287*y[49]) + y[17]*(0.16666666666666677*y[56] + 0.28867513459481275*y[50]) - y[16]*(-0.28867513459481287*y[55] - 0.16666666666666657*y[49]) - y[16]*(0.28867513459481275*y[56] - 0.16666666666666677*y[50]))/p[8] - x[1] + u[21])/p[12];
out[2] = -p[29]*x[2] + 314.15926535897933*x[3] - 314.15926535897933*y[72];
out[3] = (-p[26]*(p[28]*x[4] - 0.33333333333333331*y[66]*y[25] - 1.0*y[66]*(-0.16666666666666657*y[27] + 0.28867513459481287*y[26]) - 1.0*y[66]*(-0.16666666666666677*y[29] - 0.28867513459481275*y[28]) - 0.33333333333333331*y[60]*y[24] - 1.0*y[60]*(-0.28867513459481287*y[27] - 0.16666666666666657*y[26]) - 1.0*y[60]*(0.28867513459481275*y[29] - 0.16666666666666677*y[28]) - 0.33333333333333326*y[67]*y[27] + 0.16666666666666652*y[67]*y[29] - 0.28867513459481281*y[67]*y[28] - 0.33333333333333326*y[61]*y[26] + 0.28867513459481281*y[61]*y[29] + 0.16666666666666652*y[61]*y[28] + 0.16666666666666652*y[68]*y[27] + 0.28867513459481281*y[68]*y[26] - 0.33333333333333326*y[68]*y[29] - 0.28867513459481281*y[62]*y[27] + 0.16666666666666652*y[62]*y[26] - 0.33333333333333326*y[62]*y[28] + u[26] + y[25]*(0.16666666666666657*y[67] - 0.28867513459481287*y[61]) + y[25]*(0.16666666666666677*y[68] + 0.28867513459481275*y[62]) - y[24]*(-0.28867513459481287*y[67] - 0.16666666666666657*y[61]) - y[24]*(0.28867513459481275*y[68] - 0.16666666666666677*y[62]))/p[23] - x[3] + u[27])/p[27];
out[4] = p[30]*(1 - y[72]);

}
void g_ini_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = y[32] + 116.65548718247834*y[17] + 243.51832949342352*y[16] - 139.98658461897401*y[1] - 292.22199539210823*y[0] + 23.331097436495664*y[9] + 48.703665898684704*y[8];
out[1] = y[33] + 243.51832949342352*y[17] - 116.65548718247834*y[16] - 292.22199539210823*y[1] + 139.98658461897401*y[0] + 48.703665898684704*y[9] - 23.331097436495664*y[8];
out[2] = y[34] + 116.65548718247834*y[19] + 243.51832949342352*y[18] - 139.98658461897401*y[3] - 292.22199539210823*y[2] + 23.331097436495664*y[11] + 48.703665898684704*y[10];
out[3] = y[35] + 243.51832949342352*y[19] - 116.65548718247834*y[18] - 292.22199539210823*y[3] + 139.98658461897401*y[2] + 48.703665898684704*y[11] - 23.331097436495664*y[10];
out[4] = y[36] + 116.65548718247834*y[21] + 243.51832949342352*y[20] - 139.98658461897401*y[5] - 292.22199539210823*y[4] + 23.331097436495664*y[13] + 48.703665898684704*y[12];
out[5] = y[37] + 243.51832949342352*y[21] - 116.65548718247834*y[20] - 292.22199539210823*y[5] + 139.98658461897401*y[4] + 48.703665898684704*y[13] - 23.331097436495664*y[12];
out[6] = 116.65548718247834*y[23] + 243.51832949342352*y[22] - 139.98658461897401*y[7] - 292.22199539210823*y[6] + 23.331097436495664*y[15] + 48.703665898684704*y[14];
out[7] = 243.51832949342352*y[23] - 116.65548718247834*y[22] - 292.22199539210823*y[7] + 139.98658461897401*y[6] + 48.703665898684704*y[15] - 23.331097436495664*y[14];
out[8] = y[40] + 23.331097436495664*y[1] + 48.703665898684704*y[0] - 139.98658461897401*y[9] - 292.22199539210823*y[8] + 116.65548718247834*y[25] + 243.51832949342352*y[24];
out[9] = y[41] + 48.703665898684704*y[1] - 23.331097436495664*y[0] - 292.22199539210823*y[9] + 139.98658461897401*y[8] + 243.51832949342352*y[25] - 116.65548718247834*y[24];
out[10] = y[42] + 23.331097436495664*y[3] + 48.703665898684704*y[2] - 139.98658461897401*y[11] - 292.22199539210823*y[10] + 116.65548718247834*y[27] + 243.51832949342352*y[26];
out[11] = y[43] + 48.703665898684704*y[3] - 23.331097436495664*y[2] - 292.22199539210823*y[11] + 139.98658461897401*y[10] + 243.51832949342352*y[27] - 116.65548718247834*y[26];
out[12] = y[44] + 23.331097436495664*y[5] + 48.703665898684704*y[4] - 139.98658461897401*y[13] - 292.22199539210823*y[12] + 116.65548718247834*y[29] + 243.51832949342352*y[28];
out[13] = y[45] + 48.703665898684704*y[5] - 23.331097436495664*y[4] - 292.22199539210823*y[13] + 139.98658461897401*y[12] + 243.51832949342352*y[29] - 116.65548718247834*y[28];
out[14] = 23.331097436495664*y[7] + 48.703665898684704*y[6] - 139.98658461897401*y[15] - 292.22199539210823*y[14] + 116.65548718247834*y[31] + 243.51832949342352*y[30];
out[15] = 48.703665898684704*y[7] - 23.331097436495664*y[6] - 292.22199539210823*y[15] + 139.98658461897401*y[14] + 243.51832949342352*y[31] - 116.65548718247834*y[30];
out[16] = y[48] - 116.65548718247834*y[17] - 243.51832949342352*y[16] + 116.65548718247834*y[1] + 243.51832949342352*y[0];
out[17] = y[54] - 243.51832949342352*y[17] + 116.65548718247834*y[16] + 243.51832949342352*y[1] - 116.65548718247834*y[0];
out[18] = y[49] - 116.65548718247834*y[19] - 243.51832949342352*y[18] + 116.65548718247834*y[3] + 243.51832949342352*y[2];
out[19] = y[55] - 243.51832949342352*y[19] + 116.65548718247834*y[18] + 243.51832949342352*y[3] - 116.65548718247834*y[2];
out[20] = y[50] - 116.65548718247834*y[21] - 243.51832949342352*y[20] + 116.65548718247834*y[5] + 243.51832949342352*y[4];
out[21] = y[56] - 243.51832949342352*y[21] + 116.65548718247834*y[20] + 243.51832949342352*y[5] - 116.65548718247834*y[4];
out[22] = y[51] - 116.65548718247834*y[23] - 243.51832949342352*y[22] + 116.65548718247834*y[7] + 243.51832949342352*y[6];
out[23] = y[57] - 243.51832949342352*y[23] + 116.65548718247834*y[22] + 243.51832949342352*y[7] - 116.65548718247834*y[6];
out[24] = y[60] + 116.65548718247834*y[9] + 243.51832949342352*y[8] - 116.65548718247834*y[25] - 243.51832949342352*y[24];
out[25] = y[66] + 243.51832949342352*y[9] - 116.65548718247834*y[8] - 243.51832949342352*y[25] + 116.65548718247834*y[24];
out[26] = y[61] + 116.65548718247834*y[11] + 243.51832949342352*y[10] - 116.65548718247834*y[27] - 243.51832949342352*y[26];
out[27] = y[67] + 243.51832949342352*y[11] - 116.65548718247834*y[10] - 243.51832949342352*y[27] + 116.65548718247834*y[26];
out[28] = y[62] + 116.65548718247834*y[13] + 243.51832949342352*y[12] - 116.65548718247834*y[29] - 243.51832949342352*y[28];
out[29] = y[68] + 243.51832949342352*y[13] - 116.65548718247834*y[12] - 243.51832949342352*y[29] + 116.65548718247834*y[28];
out[30] = y[63] + 116.65548718247834*y[15] + 243.51832949342352*y[14] - 116.65548718247834*y[31] - 243.51832949342352*y[30];
out[31] = y[69] + 243.51832949342352*y[15] - 116.65548718247834*y[14] - 243.51832949342352*y[31] + 116.65548718247834*y[30];
out[32] = y[33]*y[1] - y[33]*y[7] + y[32]*y[0] - y[32]*y[6] - u[4];
out[33] = y[35]*y[3] - y[35]*y[7] + y[34]*y[2] - y[34]*y[6] - u[6];
out[34] = y[37]*y[5] - y[37]*y[7] + y[36]*y[4] - y[36]*y[6] - u[8];
out[35] = -y[33]*y[0] + y[33]*y[6] + y[32]*y[1] - y[32]*y[7] - u[5];
out[36] = -y[35]*y[2] + y[35]*y[6] + y[34]*y[3] - y[34]*y[7] - u[7];
out[37] = -y[37]*y[4] + y[37]*y[6] + y[36]*y[5] - y[36]*y[7] - u[9];
out[38] = y[32] + y[34] + y[36] + y[38];
out[39] = y[33] + y[35] + y[37] + y[39];
out[40] = y[41]*y[9] - y[41]*y[15] + y[40]*y[8] - y[40]*y[14] - u[10];
out[41] = y[43]*y[11] - y[43]*y[15] + y[42]*y[10] - y[42]*y[14] - u[12];
out[42] = y[45]*y[13] - y[45]*y[15] + y[44]*y[12] - y[44]*y[14] - u[14];
out[43] = -y[41]*y[8] + y[41]*y[14] + y[40]*y[9] - y[40]*y[15] - u[11];
out[44] = -y[43]*y[10] + y[43]*y[14] + y[42]*y[11] - y[42]*y[15] - u[13];
out[45] = -y[45]*y[12] + y[45]*y[14] + y[44]*y[13] - y[44]*y[15] - u[15];
out[46] = y[40] + y[42] + y[44] + y[46];
out[47] = y[41] + y[43] + y[45] + y[47];
out[48] = -p[1]*y[48] + 1.0*p[0]*y[54] + u[16]*cos(x[0]) - y[16] + y[22];
out[49] = -p[3]*y[49] + 1.0*p[2]*y[55] + u[17]*cos(x[0] - 2.0943951023931953) - y[18] + y[22];
out[50] = -p[5]*y[50] + 1.0*p[4]*y[56] + u[18]*cos(x[0] - 4.1887902047863905) - y[20] + y[22];
out[51] = -p[7]*y[51] + 1.0*p[6]*y[57] + y[53] - y[22];
out[52] = y[48] + y[49] + y[50] + y[51] - y[52];
out[53] = p[10]*y[52] - 1.0*p[9]*y[58] - y[53];
out[54] = -1.0*p[1]*y[54] - 1.0*p[0]*y[48] + 1.0*u[16]*sin(x[0]) - 1.0*y[17] + 1.0*y[23];
out[55] = -1.0*p[3]*y[55] - 1.0*p[2]*y[49] + 1.0*u[17]*sin(x[0] - 2.0943951023931953) - 1.0*y[19] + 1.0*y[23];
out[56] = -1.0*p[5]*y[56] - 1.0*p[4]*y[50] + 1.0*u[18]*sin(x[0] - 4.1887902047863905) - 1.0*y[21] + 1.0*y[23];
out[57] = -1.0*p[7]*y[57] - 1.0*p[6]*y[51] + 1.0*y[59] - 1.0*y[23];
out[58] = 1.0*y[54] + 1.0*y[55] + 1.0*y[56] + 1.0*y[57] - 1.0*y[58];
out[59] = 1.0*p[10]*y[58] + 1.0*p[9]*y[52] - 1.0*y[59];
out[60] = -p[16]*y[60] + 1.0*p[15]*y[66] + u[22]*cos(x[2]) - y[24] + y[30];
out[61] = -p[18]*y[61] + 1.0*p[17]*y[67] + u[23]*cos(x[2] - 2.0943951023931953) - y[26] + y[30];
out[62] = -p[20]*y[62] + 1.0*p[19]*y[68] + u[24]*cos(x[2] - 4.1887902047863905) - y[28] + y[30];
out[63] = -p[22]*y[63] + 1.0*p[21]*y[69] + y[65] - y[30];
out[64] = y[60] + y[61] + y[62] + y[63] - y[64];
out[65] = p[25]*y[64] - 1.0*p[24]*y[70] - y[65];
out[66] = -1.0*p[16]*y[66] - 1.0*p[15]*y[60] + 1.0*u[22]*sin(x[2]) - 1.0*y[25] + 1.0*y[31];
out[67] = -1.0*p[18]*y[67] - 1.0*p[17]*y[61] + 1.0*u[23]*sin(x[2] - 2.0943951023931953) - 1.0*y[27] + 1.0*y[31];
out[68] = -1.0*p[20]*y[68] - 1.0*p[19]*y[62] + 1.0*u[24]*sin(x[2] - 4.1887902047863905) - 1.0*y[29] + 1.0*y[31];
out[69] = -1.0*p[22]*y[69] - 1.0*p[21]*y[63] + 1.0*y[71] - 1.0*y[31];
out[70] = 1.0*y[66] + 1.0*y[67] + 1.0*y[68] + 1.0*y[69] - 1.0*y[70];
out[71] = 1.0*p[25]*y[70] + 1.0*p[24]*y[64] - 1.0*y[71];
out[72] = y[72] - (p[8]*x[1] + p[23]*x[3])/(p[8] + p[23]);

}
void f_run_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = -p[14]*x[0] + 314.15926535897933*x[1] - 314.15926535897933*y[72];
out[1] = (-p[11]*(p[13]*x[4] - 0.33333333333333331*y[54]*y[17] - 1.0*y[54]*(-0.16666666666666657*y[19] + 0.28867513459481287*y[18]) - 1.0*y[54]*(-0.16666666666666677*y[21] - 0.28867513459481275*y[20]) - 0.33333333333333331*y[48]*y[16] - 1.0*y[48]*(-0.28867513459481287*y[19] - 0.16666666666666657*y[18]) - 1.0*y[48]*(0.28867513459481275*y[21] - 0.16666666666666677*y[20]) - 0.33333333333333326*y[55]*y[19] + 0.16666666666666652*y[55]*y[21] - 0.28867513459481281*y[55]*y[20] - 0.33333333333333326*y[49]*y[18] + 0.28867513459481281*y[49]*y[21] + 0.16666666666666652*y[49]*y[20] + 0.16666666666666652*y[56]*y[19] + 0.28867513459481281*y[56]*y[18] - 0.33333333333333326*y[56]*y[21] - 0.28867513459481281*y[50]*y[19] + 0.16666666666666652*y[50]*y[18] - 0.33333333333333326*y[50]*y[20] + u[20] + y[17]*(0.16666666666666657*y[55] - 0.28867513459481287*y[49]) + y[17]*(0.16666666666666677*y[56] + 0.28867513459481275*y[50]) - y[16]*(-0.28867513459481287*y[55] - 0.16666666666666657*y[49]) - y[16]*(0.28867513459481275*y[56] - 0.16666666666666677*y[50]))/p[8] - x[1] + u[21])/p[12];
out[2] = -p[29]*x[2] + 314.15926535897933*x[3] - 314.15926535897933*y[72];
out[3] = (-p[26]*(p[28]*x[4] - 0.33333333333333331*y[66]*y[25] - 1.0*y[66]*(-0.16666666666666657*y[27] + 0.28867513459481287*y[26]) - 1.0*y[66]*(-0.16666666666666677*y[29] - 0.28867513459481275*y[28]) - 0.33333333333333331*y[60]*y[24] - 1.0*y[60]*(-0.28867513459481287*y[27] - 0.16666666666666657*y[26]) - 1.0*y[60]*(0.28867513459481275*y[29] - 0.16666666666666677*y[28]) - 0.33333333333333326*y[67]*y[27] + 0.16666666666666652*y[67]*y[29] - 0.28867513459481281*y[67]*y[28] - 0.33333333333333326*y[61]*y[26] + 0.28867513459481281*y[61]*y[29] + 0.16666666666666652*y[61]*y[28] + 0.16666666666666652*y[68]*y[27] + 0.28867513459481281*y[68]*y[26] - 0.33333333333333326*y[68]*y[29] - 0.28867513459481281*y[62]*y[27] + 0.16666666666666652*y[62]*y[26] - 0.33333333333333326*y[62]*y[28] + u[26] + y[25]*(0.16666666666666657*y[67] - 0.28867513459481287*y[61]) + y[25]*(0.16666666666666677*y[68] + 0.28867513459481275*y[62]) - y[24]*(-0.28867513459481287*y[67] - 0.16666666666666657*y[61]) - y[24]*(0.28867513459481275*y[68] - 0.16666666666666677*y[62]))/p[23] - x[3] + u[27])/p[27];
out[4] = p[30]*(1 - y[72]);

}
void h_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = sqrt(pow(y[1], 2) + pow(y[0], 2));
out[1] = sqrt(pow(y[3], 2) + pow(y[2], 2));
out[2] = sqrt(pow(y[5], 2) + pow(y[4], 2));
out[3] = sqrt(pow(y[7], 2) + pow(y[6], 2));
out[4] = sqrt(pow(y[9], 2) + pow(y[8], 2));
out[5] = sqrt(pow(y[11], 2) + pow(y[10], 2));
out[6] = sqrt(pow(y[13], 2) + pow(y[12], 2));
out[7] = sqrt(pow(y[15], 2) + pow(y[14], 2));
out[8] = sqrt(pow(y[17], 2) + pow(y[16], 2));
out[9] = sqrt(pow(y[19], 2) + pow(y[18], 2));
out[10] = sqrt(pow(y[21], 2) + pow(y[20], 2));
out[11] = sqrt(pow(y[23], 2) + pow(y[22], 2));
out[12] = sqrt(pow(y[25], 2) + pow(y[24], 2));
out[13] = sqrt(pow(y[27], 2) + pow(y[26], 2));
out[14] = sqrt(pow(y[29], 2) + pow(y[28], 2));
out[15] = sqrt(pow(y[31], 2) + pow(y[30], 2));
out[16] = 0.33333333333333331*y[54]*y[17] + 1.0*y[54]*(-0.16666666666666657*y[19] + 0.28867513459481287*y[18]) + 1.0*y[54]*(-0.16666666666666677*y[21] - 0.28867513459481275*y[20]) + 0.33333333333333331*y[48]*y[16] + 1.0*y[48]*(-0.28867513459481287*y[19] - 0.16666666666666657*y[18]) + 1.0*y[48]*(0.28867513459481275*y[21] - 0.16666666666666677*y[20]) + 0.33333333333333326*y[55]*y[19] - 0.16666666666666652*y[55]*y[21] + 0.28867513459481281*y[55]*y[20] + 0.33333333333333326*y[49]*y[18] - 0.28867513459481281*y[49]*y[21] - 0.16666666666666652*y[49]*y[20] - 0.16666666666666652*y[56]*y[19] - 0.28867513459481281*y[56]*y[18] + 0.33333333333333326*y[56]*y[21] + 0.28867513459481281*y[50]*y[19] - 0.16666666666666652*y[50]*y[18] + 0.33333333333333326*y[50]*y[20] - y[17]*(0.16666666666666657*y[55] - 0.28867513459481287*y[49]) - y[17]*(0.16666666666666677*y[56] + 0.28867513459481275*y[50]) + y[16]*(-0.28867513459481287*y[55] - 0.16666666666666657*y[49]) + y[16]*(0.28867513459481275*y[56] - 0.16666666666666677*y[50]);
out[17] = 0.33333333333333331*y[54]*y[17] + 1.0*y[54]*(-0.16666666666666677*y[19] - 0.28867513459481275*y[18]) + 1.0*y[54]*(-0.16666666666666657*y[21] + 0.28867513459481287*y[20]) + 0.33333333333333331*y[48]*y[16] + 1.0*y[48]*(0.28867513459481275*y[19] - 0.16666666666666677*y[18]) + 1.0*y[48]*(-0.28867513459481287*y[21] - 0.16666666666666657*y[20]) + 0.33333333333333326*y[55]*y[19] - 0.16666666666666652*y[55]*y[21] - 0.28867513459481281*y[55]*y[20] + 0.33333333333333326*y[49]*y[18] + 0.28867513459481281*y[49]*y[21] - 0.16666666666666652*y[49]*y[20] - 0.16666666666666652*y[56]*y[19] + 0.28867513459481281*y[56]*y[18] + 0.33333333333333326*y[56]*y[21] - 0.28867513459481281*y[50]*y[19] - 0.16666666666666652*y[50]*y[18] + 0.33333333333333326*y[50]*y[20] - y[17]*(0.16666666666666677*y[55] + 0.28867513459481275*y[49]) - y[17]*(0.16666666666666657*y[56] - 0.28867513459481287*y[50]) + y[16]*(0.28867513459481275*y[55] - 0.16666666666666677*y[49]) + y[16]*(-0.28867513459481287*y[56] - 0.16666666666666657*y[50]);
out[18] = 0.33333333333333331*y[54]*y[17] + 0.33333333333333331*y[54]*y[19] + 0.33333333333333331*y[54]*y[21] + 0.33333333333333331*y[48]*y[16] + 0.33333333333333331*y[48]*y[18] + 0.33333333333333331*y[48]*y[20] + 0.33333333333333331*y[55]*y[17] + 0.33333333333333331*y[55]*y[19] + 0.33333333333333331*y[55]*y[21] + 0.33333333333333331*y[49]*y[16] + 0.33333333333333331*y[49]*y[18] + 0.33333333333333331*y[49]*y[20] + 0.33333333333333331*y[56]*y[17] + 0.33333333333333331*y[56]*y[19] + 0.33333333333333331*y[56]*y[21] + 0.33333333333333331*y[50]*y[16] + 0.33333333333333331*y[50]*y[18] + 0.33333333333333331*y[50]*y[20];
out[19] = u[16];
out[20] = u[17];
out[21] = u[18];
out[22] = u[20];
out[23] = u[21];
out[24] = 0.33333333333333331*y[66]*y[25] + 1.0*y[66]*(-0.16666666666666657*y[27] + 0.28867513459481287*y[26]) + 1.0*y[66]*(-0.16666666666666677*y[29] - 0.28867513459481275*y[28]) + 0.33333333333333331*y[60]*y[24] + 1.0*y[60]*(-0.28867513459481287*y[27] - 0.16666666666666657*y[26]) + 1.0*y[60]*(0.28867513459481275*y[29] - 0.16666666666666677*y[28]) + 0.33333333333333326*y[67]*y[27] - 0.16666666666666652*y[67]*y[29] + 0.28867513459481281*y[67]*y[28] + 0.33333333333333326*y[61]*y[26] - 0.28867513459481281*y[61]*y[29] - 0.16666666666666652*y[61]*y[28] - 0.16666666666666652*y[68]*y[27] - 0.28867513459481281*y[68]*y[26] + 0.33333333333333326*y[68]*y[29] + 0.28867513459481281*y[62]*y[27] - 0.16666666666666652*y[62]*y[26] + 0.33333333333333326*y[62]*y[28] - y[25]*(0.16666666666666657*y[67] - 0.28867513459481287*y[61]) - y[25]*(0.16666666666666677*y[68] + 0.28867513459481275*y[62]) + y[24]*(-0.28867513459481287*y[67] - 0.16666666666666657*y[61]) + y[24]*(0.28867513459481275*y[68] - 0.16666666666666677*y[62]);
out[25] = 0.33333333333333331*y[66]*y[25] + 1.0*y[66]*(-0.16666666666666677*y[27] - 0.28867513459481275*y[26]) + 1.0*y[66]*(-0.16666666666666657*y[29] + 0.28867513459481287*y[28]) + 0.33333333333333331*y[60]*y[24] + 1.0*y[60]*(0.28867513459481275*y[27] - 0.16666666666666677*y[26]) + 1.0*y[60]*(-0.28867513459481287*y[29] - 0.16666666666666657*y[28]) + 0.33333333333333326*y[67]*y[27] - 0.16666666666666652*y[67]*y[29] - 0.28867513459481281*y[67]*y[28] + 0.33333333333333326*y[61]*y[26] + 0.28867513459481281*y[61]*y[29] - 0.16666666666666652*y[61]*y[28] - 0.16666666666666652*y[68]*y[27] + 0.28867513459481281*y[68]*y[26] + 0.33333333333333326*y[68]*y[29] - 0.28867513459481281*y[62]*y[27] - 0.16666666666666652*y[62]*y[26] + 0.33333333333333326*y[62]*y[28] - y[25]*(0.16666666666666677*y[67] + 0.28867513459481275*y[61]) - y[25]*(0.16666666666666657*y[68] - 0.28867513459481287*y[62]) + y[24]*(0.28867513459481275*y[67] - 0.16666666666666677*y[61]) + y[24]*(-0.28867513459481287*y[68] - 0.16666666666666657*y[62]);
out[26] = 0.33333333333333331*y[66]*y[25] + 0.33333333333333331*y[66]*y[27] + 0.33333333333333331*y[66]*y[29] + 0.33333333333333331*y[60]*y[24] + 0.33333333333333331*y[60]*y[26] + 0.33333333333333331*y[60]*y[28] + 0.33333333333333331*y[67]*y[25] + 0.33333333333333331*y[67]*y[27] + 0.33333333333333331*y[67]*y[29] + 0.33333333333333331*y[61]*y[24] + 0.33333333333333331*y[61]*y[26] + 0.33333333333333331*y[61]*y[28] + 0.33333333333333331*y[68]*y[25] + 0.33333333333333331*y[68]*y[27] + 0.33333333333333331*y[68]*y[29] + 0.33333333333333331*y[62]*y[24] + 0.33333333333333331*y[62]*y[26] + 0.33333333333333331*y[62]*y[28];
out[27] = u[22];
out[28] = u[23];
out[29] = u[24];
out[30] = u[26];
out[31] = u[27];

}
void g_run_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = y[32] + 116.65548718247834*y[17] + 243.51832949342352*y[16] - 139.98658461897401*y[1] - 292.22199539210823*y[0] + 23.331097436495664*y[9] + 48.703665898684704*y[8];
out[1] = y[33] + 243.51832949342352*y[17] - 116.65548718247834*y[16] - 292.22199539210823*y[1] + 139.98658461897401*y[0] + 48.703665898684704*y[9] - 23.331097436495664*y[8];
out[2] = y[34] + 116.65548718247834*y[19] + 243.51832949342352*y[18] - 139.98658461897401*y[3] - 292.22199539210823*y[2] + 23.331097436495664*y[11] + 48.703665898684704*y[10];
out[3] = y[35] + 243.51832949342352*y[19] - 116.65548718247834*y[18] - 292.22199539210823*y[3] + 139.98658461897401*y[2] + 48.703665898684704*y[11] - 23.331097436495664*y[10];
out[4] = y[36] + 116.65548718247834*y[21] + 243.51832949342352*y[20] - 139.98658461897401*y[5] - 292.22199539210823*y[4] + 23.331097436495664*y[13] + 48.703665898684704*y[12];
out[5] = y[37] + 243.51832949342352*y[21] - 116.65548718247834*y[20] - 292.22199539210823*y[5] + 139.98658461897401*y[4] + 48.703665898684704*y[13] - 23.331097436495664*y[12];
out[6] = 116.65548718247834*y[23] + 243.51832949342352*y[22] - 139.98658461897401*y[7] - 292.22199539210823*y[6] + 23.331097436495664*y[15] + 48.703665898684704*y[14];
out[7] = 243.51832949342352*y[23] - 116.65548718247834*y[22] - 292.22199539210823*y[7] + 139.98658461897401*y[6] + 48.703665898684704*y[15] - 23.331097436495664*y[14];
out[8] = y[40] + 23.331097436495664*y[1] + 48.703665898684704*y[0] - 139.98658461897401*y[9] - 292.22199539210823*y[8] + 116.65548718247834*y[25] + 243.51832949342352*y[24];
out[9] = y[41] + 48.703665898684704*y[1] - 23.331097436495664*y[0] - 292.22199539210823*y[9] + 139.98658461897401*y[8] + 243.51832949342352*y[25] - 116.65548718247834*y[24];
out[10] = y[42] + 23.331097436495664*y[3] + 48.703665898684704*y[2] - 139.98658461897401*y[11] - 292.22199539210823*y[10] + 116.65548718247834*y[27] + 243.51832949342352*y[26];
out[11] = y[43] + 48.703665898684704*y[3] - 23.331097436495664*y[2] - 292.22199539210823*y[11] + 139.98658461897401*y[10] + 243.51832949342352*y[27] - 116.65548718247834*y[26];
out[12] = y[44] + 23.331097436495664*y[5] + 48.703665898684704*y[4] - 139.98658461897401*y[13] - 292.22199539210823*y[12] + 116.65548718247834*y[29] + 243.51832949342352*y[28];
out[13] = y[45] + 48.703665898684704*y[5] - 23.331097436495664*y[4] - 292.22199539210823*y[13] + 139.98658461897401*y[12] + 243.51832949342352*y[29] - 116.65548718247834*y[28];
out[14] = 23.331097436495664*y[7] + 48.703665898684704*y[6] - 139.98658461897401*y[15] - 292.22199539210823*y[14] + 116.65548718247834*y[31] + 243.51832949342352*y[30];
out[15] = 48.703665898684704*y[7] - 23.331097436495664*y[6] - 292.22199539210823*y[15] + 139.98658461897401*y[14] + 243.51832949342352*y[31] - 116.65548718247834*y[30];
out[16] = y[48] - 116.65548718247834*y[17] - 243.51832949342352*y[16] + 116.65548718247834*y[1] + 243.51832949342352*y[0];
out[17] = y[54] - 243.51832949342352*y[17] + 116.65548718247834*y[16] + 243.51832949342352*y[1] - 116.65548718247834*y[0];
out[18] = y[49] - 116.65548718247834*y[19] - 243.51832949342352*y[18] + 116.65548718247834*y[3] + 243.51832949342352*y[2];
out[19] = y[55] - 243.51832949342352*y[19] + 116.65548718247834*y[18] + 243.51832949342352*y[3] - 116.65548718247834*y[2];
out[20] = y[50] - 116.65548718247834*y[21] - 243.51832949342352*y[20] + 116.65548718247834*y[5] + 243.51832949342352*y[4];
out[21] = y[56] - 243.51832949342352*y[21] + 116.65548718247834*y[20] + 243.51832949342352*y[5] - 116.65548718247834*y[4];
out[22] = y[51] - 116.65548718247834*y[23] - 243.51832949342352*y[22] + 116.65548718247834*y[7] + 243.51832949342352*y[6];
out[23] = y[57] - 243.51832949342352*y[23] + 116.65548718247834*y[22] + 243.51832949342352*y[7] - 116.65548718247834*y[6];
out[24] = y[60] + 116.65548718247834*y[9] + 243.51832949342352*y[8] - 116.65548718247834*y[25] - 243.51832949342352*y[24];
out[25] = y[66] + 243.51832949342352*y[9] - 116.65548718247834*y[8] - 243.51832949342352*y[25] + 116.65548718247834*y[24];
out[26] = y[61] + 116.65548718247834*y[11] + 243.51832949342352*y[10] - 116.65548718247834*y[27] - 243.51832949342352*y[26];
out[27] = y[67] + 243.51832949342352*y[11] - 116.65548718247834*y[10] - 243.51832949342352*y[27] + 116.65548718247834*y[26];
out[28] = y[62] + 116.65548718247834*y[13] + 243.51832949342352*y[12] - 116.65548718247834*y[29] - 243.51832949342352*y[28];
out[29] = y[68] + 243.51832949342352*y[13] - 116.65548718247834*y[12] - 243.51832949342352*y[29] + 116.65548718247834*y[28];
out[30] = y[63] + 116.65548718247834*y[15] + 243.51832949342352*y[14] - 116.65548718247834*y[31] - 243.51832949342352*y[30];
out[31] = y[69] + 243.51832949342352*y[15] - 116.65548718247834*y[14] - 243.51832949342352*y[31] + 116.65548718247834*y[30];
out[32] = y[33]*y[1] - y[33]*y[7] + y[32]*y[0] - y[32]*y[6] - u[4];
out[33] = y[35]*y[3] - y[35]*y[7] + y[34]*y[2] - y[34]*y[6] - u[6];
out[34] = y[37]*y[5] - y[37]*y[7] + y[36]*y[4] - y[36]*y[6] - u[8];
out[35] = -y[33]*y[0] + y[33]*y[6] + y[32]*y[1] - y[32]*y[7] - u[5];
out[36] = -y[35]*y[2] + y[35]*y[6] + y[34]*y[3] - y[34]*y[7] - u[7];
out[37] = -y[37]*y[4] + y[37]*y[6] + y[36]*y[5] - y[36]*y[7] - u[9];
out[38] = y[32] + y[34] + y[36] + y[38];
out[39] = y[33] + y[35] + y[37] + y[39];
out[40] = y[41]*y[9] - y[41]*y[15] + y[40]*y[8] - y[40]*y[14] - u[10];
out[41] = y[43]*y[11] - y[43]*y[15] + y[42]*y[10] - y[42]*y[14] - u[12];
out[42] = y[45]*y[13] - y[45]*y[15] + y[44]*y[12] - y[44]*y[14] - u[14];
out[43] = -y[41]*y[8] + y[41]*y[14] + y[40]*y[9] - y[40]*y[15] - u[11];
out[44] = -y[43]*y[10] + y[43]*y[14] + y[42]*y[11] - y[42]*y[15] - u[13];
out[45] = -y[45]*y[12] + y[45]*y[14] + y[44]*y[13] - y[44]*y[15] - u[15];
out[46] = y[40] + y[42] + y[44] + y[46];
out[47] = y[41] + y[43] + y[45] + y[47];
out[48] = -p[1]*y[48] + 1.0*p[0]*y[54] + u[16]*cos(x[0]) - y[16] + y[22];
out[49] = -p[3]*y[49] + 1.0*p[2]*y[55] + u[17]*cos(x[0] - 2.0943951023931953) - y[18] + y[22];
out[50] = -p[5]*y[50] + 1.0*p[4]*y[56] + u[18]*cos(x[0] - 4.1887902047863905) - y[20] + y[22];
out[51] = -p[7]*y[51] + 1.0*p[6]*y[57] + y[53] - y[22];
out[52] = y[48] + y[49] + y[50] + y[51] - y[52];
out[53] = p[10]*y[52] - 1.0*p[9]*y[58] - y[53];
out[54] = -1.0*p[1]*y[54] - 1.0*p[0]*y[48] + 1.0*u[16]*sin(x[0]) - 1.0*y[17] + 1.0*y[23];
out[55] = -1.0*p[3]*y[55] - 1.0*p[2]*y[49] + 1.0*u[17]*sin(x[0] - 2.0943951023931953) - 1.0*y[19] + 1.0*y[23];
out[56] = -1.0*p[5]*y[56] - 1.0*p[4]*y[50] + 1.0*u[18]*sin(x[0] - 4.1887902047863905) - 1.0*y[21] + 1.0*y[23];
out[57] = -1.0*p[7]*y[57] - 1.0*p[6]*y[51] + 1.0*y[59] - 1.0*y[23];
out[58] = 1.0*y[54] + 1.0*y[55] + 1.0*y[56] + 1.0*y[57] - 1.0*y[58];
out[59] = 1.0*p[10]*y[58] + 1.0*p[9]*y[52] - 1.0*y[59];
out[60] = -p[16]*y[60] + 1.0*p[15]*y[66] + u[22]*cos(x[2]) - y[24] + y[30];
out[61] = -p[18]*y[61] + 1.0*p[17]*y[67] + u[23]*cos(x[2] - 2.0943951023931953) - y[26] + y[30];
out[62] = -p[20]*y[62] + 1.0*p[19]*y[68] + u[24]*cos(x[2] - 4.1887902047863905) - y[28] + y[30];
out[63] = -p[22]*y[63] + 1.0*p[21]*y[69] + y[65] - y[30];
out[64] = y[60] + y[61] + y[62] + y[63] - y[64];
out[65] = p[25]*y[64] - 1.0*p[24]*y[70] - y[65];
out[66] = -1.0*p[16]*y[66] - 1.0*p[15]*y[60] + 1.0*u[22]*sin(x[2]) - 1.0*y[25] + 1.0*y[31];
out[67] = -1.0*p[18]*y[67] - 1.0*p[17]*y[61] + 1.0*u[23]*sin(x[2] - 2.0943951023931953) - 1.0*y[27] + 1.0*y[31];
out[68] = -1.0*p[20]*y[68] - 1.0*p[19]*y[62] + 1.0*u[24]*sin(x[2] - 4.1887902047863905) - 1.0*y[29] + 1.0*y[31];
out[69] = -1.0*p[22]*y[69] - 1.0*p[21]*y[63] + 1.0*y[71] - 1.0*y[31];
out[70] = 1.0*y[66] + 1.0*y[67] + 1.0*y[68] + 1.0*y[69] - 1.0*y[70];
out[71] = 1.0*p[25]*y[70] + 1.0*p[24]*y[64] - 1.0*y[71];
out[72] = y[72] - (p[8]*x[1] + p[23]*x[3])/(p[8] + p[23]);

}
void de_jac_ini_xy_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[99] = -p[11]*(-0.33333333333333331*y[48] + 0.28867513459481287*y[55] + 0.16666666666666657*y[49] - 0.28867513459481275*y[56] + 0.16666666666666677*y[50])/(p[8]*p[12]);
out[100] = -p[11]*(-0.33333333333333331*y[54] + 0.16666666666666657*y[55] - 0.28867513459481287*y[49] + 0.16666666666666677*y[56] + 0.28867513459481275*y[50])/(p[8]*p[12]);
out[101] = -p[11]*(-0.28867513459481287*y[54] + 0.16666666666666657*y[48] - 0.33333333333333326*y[49] + 0.28867513459481281*y[56] + 0.16666666666666652*y[50])/(p[8]*p[12]);
out[102] = -p[11]*(0.16666666666666657*y[54] + 0.28867513459481287*y[48] - 0.33333333333333326*y[55] + 0.16666666666666652*y[56] - 0.28867513459481281*y[50])/(p[8]*p[12]);
out[103] = -p[11]*(0.28867513459481275*y[54] + 0.16666666666666677*y[48] - 0.28867513459481281*y[55] + 0.16666666666666652*y[49] - 0.33333333333333326*y[50])/(p[8]*p[12]);
out[104] = -p[11]*(0.16666666666666677*y[54] - 0.28867513459481275*y[48] + 0.16666666666666652*y[55] + 0.28867513459481281*y[49] - 0.33333333333333326*y[56])/(p[8]*p[12]);
out[131] = -p[11]*(-0.33333333333333331*y[16] + 0.28867513459481287*y[19] + 0.16666666666666657*y[18] - 0.28867513459481275*y[21] + 0.16666666666666677*y[20])/(p[8]*p[12]);
out[132] = -p[11]*(-0.28867513459481287*y[17] + 0.16666666666666657*y[16] - 0.33333333333333326*y[18] + 0.28867513459481281*y[21] + 0.16666666666666652*y[20])/(p[8]*p[12]);
out[133] = -p[11]*(0.28867513459481275*y[17] + 0.16666666666666677*y[16] - 0.28867513459481281*y[19] + 0.16666666666666652*y[18] - 0.33333333333333326*y[20])/(p[8]*p[12]);
out[137] = -p[11]*(-0.33333333333333331*y[17] + 0.16666666666666657*y[19] - 0.28867513459481287*y[18] + 0.16666666666666677*y[21] + 0.28867513459481275*y[20])/(p[8]*p[12]);
out[138] = -p[11]*(0.16666666666666657*y[17] + 0.28867513459481287*y[16] - 0.33333333333333326*y[19] + 0.16666666666666652*y[21] - 0.28867513459481281*y[20])/(p[8]*p[12]);
out[139] = -p[11]*(0.16666666666666677*y[17] - 0.28867513459481275*y[16] + 0.16666666666666652*y[19] + 0.28867513459481281*y[18] - 0.33333333333333326*y[21])/(p[8]*p[12]);
out[263] = -p[26]*(-0.33333333333333331*y[60] + 0.28867513459481287*y[67] + 0.16666666666666657*y[61] - 0.28867513459481275*y[68] + 0.16666666666666677*y[62])/(p[23]*p[27]);
out[264] = -p[26]*(-0.33333333333333331*y[66] + 0.16666666666666657*y[67] - 0.28867513459481287*y[61] + 0.16666666666666677*y[68] + 0.28867513459481275*y[62])/(p[23]*p[27]);
out[265] = -p[26]*(-0.28867513459481287*y[66] + 0.16666666666666657*y[60] - 0.33333333333333326*y[61] + 0.28867513459481281*y[68] + 0.16666666666666652*y[62])/(p[23]*p[27]);
out[266] = -p[26]*(0.16666666666666657*y[66] + 0.28867513459481287*y[60] - 0.33333333333333326*y[67] + 0.16666666666666652*y[68] - 0.28867513459481281*y[62])/(p[23]*p[27]);
out[267] = -p[26]*(0.28867513459481275*y[66] + 0.16666666666666677*y[60] - 0.28867513459481281*y[67] + 0.16666666666666652*y[61] - 0.33333333333333326*y[62])/(p[23]*p[27]);
out[268] = -p[26]*(0.16666666666666677*y[66] - 0.28867513459481275*y[60] + 0.16666666666666652*y[67] + 0.28867513459481281*y[61] - 0.33333333333333326*y[68])/(p[23]*p[27]);
out[299] = -p[26]*(-0.33333333333333331*y[24] + 0.28867513459481287*y[27] + 0.16666666666666657*y[26] - 0.28867513459481275*y[29] + 0.16666666666666677*y[28])/(p[23]*p[27]);
out[300] = -p[26]*(-0.28867513459481287*y[25] + 0.16666666666666657*y[24] - 0.33333333333333326*y[26] + 0.28867513459481281*y[29] + 0.16666666666666652*y[28])/(p[23]*p[27]);
out[301] = -p[26]*(0.28867513459481275*y[25] + 0.16666666666666677*y[24] - 0.28867513459481281*y[27] + 0.16666666666666652*y[26] - 0.33333333333333326*y[28])/(p[23]*p[27]);
out[305] = -p[26]*(-0.33333333333333331*y[25] + 0.16666666666666657*y[27] - 0.28867513459481287*y[26] + 0.16666666666666677*y[29] + 0.28867513459481275*y[28])/(p[23]*p[27]);
out[306] = -p[26]*(0.16666666666666657*y[25] + 0.28867513459481287*y[24] - 0.33333333333333326*y[27] + 0.16666666666666652*y[29] - 0.28867513459481281*y[28])/(p[23]*p[27]);
out[307] = -p[26]*(0.16666666666666677*y[25] - 0.28867513459481275*y[24] + 0.16666666666666652*y[27] + 0.28867513459481281*y[26] - 0.33333333333333326*y[29])/(p[23]*p[27]);
out[2891] = y[32];
out[2892] = y[33];
out[2897] = -y[32];
out[2898] = -y[33];
out[2923] = y[0] - y[6];
out[2924] = y[1] - y[7];
out[2971] = y[34];
out[2972] = y[35];
out[2975] = -y[34];
out[2976] = -y[35];
out[3003] = y[2] - y[6];
out[3004] = y[3] - y[7];
out[3051] = y[36];
out[3052] = y[37];
out[3053] = -y[36];
out[3054] = -y[37];
out[3083] = y[4] - y[6];
out[3084] = y[5] - y[7];
out[3125] = -y[33];
out[3126] = y[32];
out[3131] = y[33];
out[3132] = -y[32];
out[3157] = y[1] - y[7];
out[3158] = -y[0] + y[6];
out[3205] = -y[35];
out[3206] = y[34];
out[3209] = y[35];
out[3210] = -y[34];
out[3237] = y[3] - y[7];
out[3238] = -y[2] + y[6];
out[3285] = -y[37];
out[3286] = y[36];
out[3287] = y[37];
out[3288] = -y[36];
out[3317] = y[5] - y[7];
out[3318] = -y[4] + y[6];
out[3523] = y[40];
out[3524] = y[41];
out[3529] = -y[40];
out[3530] = -y[41];
out[3555] = y[8] - y[14];
out[3556] = y[9] - y[15];
out[3603] = y[42];
out[3604] = y[43];
out[3607] = -y[42];
out[3608] = -y[43];
out[3635] = y[10] - y[14];
out[3636] = y[11] - y[15];
out[3683] = y[44];
out[3684] = y[45];
out[3685] = -y[44];
out[3686] = -y[45];
out[3715] = y[12] - y[14];
out[3716] = y[13] - y[15];
out[3757] = -y[41];
out[3758] = y[40];
out[3763] = y[41];
out[3764] = -y[40];
out[3789] = y[9] - y[15];
out[3790] = -y[8] + y[14];
out[3837] = -y[43];
out[3838] = y[42];
out[3841] = y[43];
out[3842] = -y[42];
out[3869] = y[11] - y[15];
out[3870] = -y[10] + y[14];
out[3917] = -y[45];
out[3918] = y[44];
out[3919] = y[45];
out[3920] = -y[44];
out[3949] = y[13] - y[15];
out[3950] = -y[12] + y[14];
out[4134] = -u[16]*sin(x[0]);
out[4212] = -u[17]*sin(x[0] - 2.0943951023931953);
out[4290] = -u[18]*sin(x[0] - 4.1887902047863905);
out[4602] = 1.0*u[16]*cos(x[0]);
out[4680] = 1.0*u[17]*cos(x[0] - 2.0943951023931953);
out[4758] = 1.0*u[18]*cos(x[0] - 4.1887902047863905);
out[5072] = -u[22]*sin(x[2]);
out[5150] = -u[23]*sin(x[2] - 2.0943951023931953);
out[5228] = -u[24]*sin(x[2] - 4.1887902047863905);
out[5540] = 1.0*u[22]*cos(x[2]);
out[5618] = 1.0*u[23]*cos(x[2] - 2.0943951023931953);
out[5696] = 1.0*u[24]*cos(x[2] - 4.1887902047863905);

}

void de_jac_ini_up_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = -p[14];
out[79] = -1/p[12];
out[82] = -p[11]*p[13]/(p[8]*p[12]);
out[158] = -p[29];
out[237] = -1/p[27];
out[238] = -p[26]*p[28]/(p[23]*p[27]);
out[389] = -p[30];
out[4187] = -p[1];
out[4193] = 1.0*p[0];
out[4266] = -p[3];
out[4272] = 1.0*p[2];
out[4345] = -p[5];
out[4351] = 1.0*p[4];
out[4424] = -p[7];
out[4430] = 1.0*p[6];
out[4581] = p[10];
out[4587] = -1.0*p[9];
out[4655] = -1.0*p[0];
out[4661] = -1.0*p[1];
out[4734] = -1.0*p[2];
out[4740] = -1.0*p[3];
out[4813] = -1.0*p[4];
out[4819] = -1.0*p[5];
out[4892] = -1.0*p[6];
out[4898] = -1.0*p[7];
out[5049] = 1.0*p[9];
out[5055] = 1.0*p[10];
out[5135] = -p[16];
out[5141] = 1.0*p[15];
out[5214] = -p[18];
out[5220] = 1.0*p[17];
out[5293] = -p[20];
out[5299] = 1.0*p[19];
out[5372] = -p[22];
out[5378] = 1.0*p[21];
out[5529] = p[25];
out[5535] = -1.0*p[24];
out[5603] = -1.0*p[15];
out[5609] = -1.0*p[16];
out[5682] = -1.0*p[17];
out[5688] = -1.0*p[18];
out[5761] = -1.0*p[19];
out[5767] = -1.0*p[20];
out[5840] = -1.0*p[21];
out[5846] = -1.0*p[22];
out[5997] = 1.0*p[24];
out[6003] = 1.0*p[25];
out[6007] = -p[8]/(p[8] + p[23]);
out[6009] = -p[23]/(p[8] + p[23]);

}

void de_jac_ini_num_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[1] = 314.15926535897933;
out[77] = -314.15926535897933;
out[159] = 314.15926535897933;
out[233] = -314.15926535897933;
out[395] = -292.22199539210823;
out[396] = -139.98658461897401;
out[403] = 48.703665898684704;
out[404] = 23.331097436495664;
out[411] = 243.51832949342352;
out[412] = 116.65548718247834;
out[427] = 1;
out[473] = 139.98658461897401;
out[474] = -292.22199539210823;
out[481] = -23.331097436495664;
out[482] = 48.703665898684704;
out[489] = -116.65548718247834;
out[490] = 243.51832949342352;
out[506] = 1;
out[553] = -292.22199539210823;
out[554] = -139.98658461897401;
out[561] = 48.703665898684704;
out[562] = 23.331097436495664;
out[569] = 243.51832949342352;
out[570] = 116.65548718247834;
out[585] = 1;
out[631] = 139.98658461897401;
out[632] = -292.22199539210823;
out[639] = -23.331097436495664;
out[640] = 48.703665898684704;
out[647] = -116.65548718247834;
out[648] = 243.51832949342352;
out[664] = 1;
out[711] = -292.22199539210823;
out[712] = -139.98658461897401;
out[719] = 48.703665898684704;
out[720] = 23.331097436495664;
out[727] = 243.51832949342352;
out[728] = 116.65548718247834;
out[743] = 1;
out[789] = 139.98658461897401;
out[790] = -292.22199539210823;
out[797] = -23.331097436495664;
out[798] = 48.703665898684704;
out[805] = -116.65548718247834;
out[806] = 243.51832949342352;
out[822] = 1;
out[869] = -292.22199539210823;
out[870] = -139.98658461897401;
out[877] = 48.703665898684704;
out[878] = 23.331097436495664;
out[885] = 243.51832949342352;
out[886] = 116.65548718247834;
out[947] = 139.98658461897401;
out[948] = -292.22199539210823;
out[955] = -23.331097436495664;
out[956] = 48.703665898684704;
out[963] = -116.65548718247834;
out[964] = 243.51832949342352;
out[1019] = 48.703665898684704;
out[1020] = 23.331097436495664;
out[1027] = -292.22199539210823;
out[1028] = -139.98658461897401;
out[1043] = 243.51832949342352;
out[1044] = 116.65548718247834;
out[1059] = 1;
out[1097] = -23.331097436495664;
out[1098] = 48.703665898684704;
out[1105] = 139.98658461897401;
out[1106] = -292.22199539210823;
out[1121] = -116.65548718247834;
out[1122] = 243.51832949342352;
out[1138] = 1;
out[1177] = 48.703665898684704;
out[1178] = 23.331097436495664;
out[1185] = -292.22199539210823;
out[1186] = -139.98658461897401;
out[1201] = 243.51832949342352;
out[1202] = 116.65548718247834;
out[1217] = 1;
out[1255] = -23.331097436495664;
out[1256] = 48.703665898684704;
out[1263] = 139.98658461897401;
out[1264] = -292.22199539210823;
out[1279] = -116.65548718247834;
out[1280] = 243.51832949342352;
out[1296] = 1;
out[1335] = 48.703665898684704;
out[1336] = 23.331097436495664;
out[1343] = -292.22199539210823;
out[1344] = -139.98658461897401;
out[1359] = 243.51832949342352;
out[1360] = 116.65548718247834;
out[1375] = 1;
out[1413] = -23.331097436495664;
out[1414] = 48.703665898684704;
out[1421] = 139.98658461897401;
out[1422] = -292.22199539210823;
out[1437] = -116.65548718247834;
out[1438] = 243.51832949342352;
out[1454] = 1;
out[1493] = 48.703665898684704;
out[1494] = 23.331097436495664;
out[1501] = -292.22199539210823;
out[1502] = -139.98658461897401;
out[1517] = 243.51832949342352;
out[1518] = 116.65548718247834;
out[1571] = -23.331097436495664;
out[1572] = 48.703665898684704;
out[1579] = 139.98658461897401;
out[1580] = -292.22199539210823;
out[1595] = -116.65548718247834;
out[1596] = 243.51832949342352;
out[1643] = 243.51832949342352;
out[1644] = 116.65548718247834;
out[1659] = -243.51832949342352;
out[1660] = -116.65548718247834;
out[1691] = 1;
out[1721] = -116.65548718247834;
out[1722] = 243.51832949342352;
out[1737] = 116.65548718247834;
out[1738] = -243.51832949342352;
out[1775] = 1;
out[1801] = 243.51832949342352;
out[1802] = 116.65548718247834;
out[1817] = -243.51832949342352;
out[1818] = -116.65548718247834;
out[1848] = 1;
out[1879] = -116.65548718247834;
out[1880] = 243.51832949342352;
out[1895] = 116.65548718247834;
out[1896] = -243.51832949342352;
out[1932] = 1;
out[1959] = 243.51832949342352;
out[1960] = 116.65548718247834;
out[1975] = -243.51832949342352;
out[1976] = -116.65548718247834;
out[2005] = 1;
out[2037] = -116.65548718247834;
out[2038] = 243.51832949342352;
out[2053] = 116.65548718247834;
out[2054] = -243.51832949342352;
out[2089] = 1;
out[2117] = 243.51832949342352;
out[2118] = 116.65548718247834;
out[2133] = -243.51832949342352;
out[2134] = -116.65548718247834;
out[2162] = 1;
out[2195] = -116.65548718247834;
out[2196] = 243.51832949342352;
out[2211] = 116.65548718247834;
out[2212] = -243.51832949342352;
out[2246] = 1;
out[2275] = 243.51832949342352;
out[2276] = 116.65548718247834;
out[2291] = -243.51832949342352;
out[2292] = -116.65548718247834;
out[2327] = 1;
out[2353] = -116.65548718247834;
out[2354] = 243.51832949342352;
out[2369] = 116.65548718247834;
out[2370] = -243.51832949342352;
out[2411] = 1;
out[2433] = 243.51832949342352;
out[2434] = 116.65548718247834;
out[2449] = -243.51832949342352;
out[2450] = -116.65548718247834;
out[2484] = 1;
out[2511] = -116.65548718247834;
out[2512] = 243.51832949342352;
out[2527] = 116.65548718247834;
out[2528] = -243.51832949342352;
out[2568] = 1;
out[2591] = 243.51832949342352;
out[2592] = 116.65548718247834;
out[2607] = -243.51832949342352;
out[2608] = -116.65548718247834;
out[2641] = 1;
out[2669] = -116.65548718247834;
out[2670] = 243.51832949342352;
out[2685] = 116.65548718247834;
out[2686] = -243.51832949342352;
out[2725] = 1;
out[2749] = 243.51832949342352;
out[2750] = 116.65548718247834;
out[2765] = -243.51832949342352;
out[2766] = -116.65548718247834;
out[2798] = 1;
out[2827] = -116.65548718247834;
out[2828] = 243.51832949342352;
out[2843] = 116.65548718247834;
out[2844] = -243.51832949342352;
out[2882] = 1;
out[3391] = 1;
out[3393] = 1;
out[3395] = 1;
out[3397] = 1;
out[3470] = 1;
out[3472] = 1;
out[3474] = 1;
out[3476] = 1;
out[4023] = 1;
out[4025] = 1;
out[4027] = 1;
out[4029] = 1;
out[4102] = 1;
out[4104] = 1;
out[4106] = 1;
out[4108] = 1;
out[4155] = -1;
out[4161] = 1;
out[4235] = -1;
out[4239] = 1;
out[4315] = -1;
out[4317] = 1;
out[4395] = -1;
out[4426] = 1;
out[4499] = 1;
out[4500] = 1;
out[4501] = 1;
out[4502] = 1;
out[4503] = -1;
out[4582] = -1;
out[4624] = -1.0;
out[4630] = 1.0;
out[4704] = -1.0;
out[4708] = 1.0;
out[4784] = -1.0;
out[4786] = 1.0;
out[4864] = -1.0;
out[4900] = 1.0;
out[4973] = 1.0;
out[4974] = 1.0;
out[4975] = 1.0;
out[4976] = 1.0;
out[4977] = -1.0;
out[5056] = -1.0;
out[5099] = -1;
out[5105] = 1;
out[5179] = -1;
out[5183] = 1;
out[5259] = -1;
out[5261] = 1;
out[5339] = -1;
out[5374] = 1;
out[5447] = 1;
out[5448] = 1;
out[5449] = 1;
out[5450] = 1;
out[5451] = -1;
out[5530] = -1;
out[5568] = -1.0;
out[5574] = 1.0;
out[5648] = -1.0;
out[5652] = 1.0;
out[5728] = -1.0;
out[5730] = 1.0;
out[5808] = -1.0;
out[5848] = 1.0;
out[5921] = 1.0;
out[5922] = 1.0;
out[5923] = 1.0;
out[5924] = 1.0;
out[5925] = -1.0;
out[6004] = -1.0;
out[6083] = 1;

}

void sp_jac_ini_xy_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[5] = -p[11]*(-0.33333333333333331*y[48] + 0.28867513459481287*y[55] + 0.16666666666666657*y[49] - 0.28867513459481275*y[56] + 0.16666666666666677*y[50])/(p[8]*p[12]);
out[6] = -p[11]*(-0.33333333333333331*y[54] + 0.16666666666666657*y[55] - 0.28867513459481287*y[49] + 0.16666666666666677*y[56] + 0.28867513459481275*y[50])/(p[8]*p[12]);
out[7] = -p[11]*(-0.28867513459481287*y[54] + 0.16666666666666657*y[48] - 0.33333333333333326*y[49] + 0.28867513459481281*y[56] + 0.16666666666666652*y[50])/(p[8]*p[12]);
out[8] = -p[11]*(0.16666666666666657*y[54] + 0.28867513459481287*y[48] - 0.33333333333333326*y[55] + 0.16666666666666652*y[56] - 0.28867513459481281*y[50])/(p[8]*p[12]);
out[9] = -p[11]*(0.28867513459481275*y[54] + 0.16666666666666677*y[48] - 0.28867513459481281*y[55] + 0.16666666666666652*y[49] - 0.33333333333333326*y[50])/(p[8]*p[12]);
out[10] = -p[11]*(0.16666666666666677*y[54] - 0.28867513459481275*y[48] + 0.16666666666666652*y[55] + 0.28867513459481281*y[49] - 0.33333333333333326*y[56])/(p[8]*p[12]);
out[11] = -p[11]*(-0.33333333333333331*y[16] + 0.28867513459481287*y[19] + 0.16666666666666657*y[18] - 0.28867513459481275*y[21] + 0.16666666666666677*y[20])/(p[8]*p[12]);
out[12] = -p[11]*(-0.28867513459481287*y[17] + 0.16666666666666657*y[16] - 0.33333333333333326*y[18] + 0.28867513459481281*y[21] + 0.16666666666666652*y[20])/(p[8]*p[12]);
out[13] = -p[11]*(0.28867513459481275*y[17] + 0.16666666666666677*y[16] - 0.28867513459481281*y[19] + 0.16666666666666652*y[18] - 0.33333333333333326*y[20])/(p[8]*p[12]);
out[14] = -p[11]*(-0.33333333333333331*y[17] + 0.16666666666666657*y[19] - 0.28867513459481287*y[18] + 0.16666666666666677*y[21] + 0.28867513459481275*y[20])/(p[8]*p[12]);
out[15] = -p[11]*(0.16666666666666657*y[17] + 0.28867513459481287*y[16] - 0.33333333333333326*y[19] + 0.16666666666666652*y[21] - 0.28867513459481281*y[20])/(p[8]*p[12]);
out[16] = -p[11]*(0.16666666666666677*y[17] - 0.28867513459481275*y[16] + 0.16666666666666652*y[19] + 0.28867513459481281*y[18] - 0.33333333333333326*y[21])/(p[8]*p[12]);
out[22] = -p[26]*(-0.33333333333333331*y[60] + 0.28867513459481287*y[67] + 0.16666666666666657*y[61] - 0.28867513459481275*y[68] + 0.16666666666666677*y[62])/(p[23]*p[27]);
out[23] = -p[26]*(-0.33333333333333331*y[66] + 0.16666666666666657*y[67] - 0.28867513459481287*y[61] + 0.16666666666666677*y[68] + 0.28867513459481275*y[62])/(p[23]*p[27]);
out[24] = -p[26]*(-0.28867513459481287*y[66] + 0.16666666666666657*y[60] - 0.33333333333333326*y[61] + 0.28867513459481281*y[68] + 0.16666666666666652*y[62])/(p[23]*p[27]);
out[25] = -p[26]*(0.16666666666666657*y[66] + 0.28867513459481287*y[60] - 0.33333333333333326*y[67] + 0.16666666666666652*y[68] - 0.28867513459481281*y[62])/(p[23]*p[27]);
out[26] = -p[26]*(0.28867513459481275*y[66] + 0.16666666666666677*y[60] - 0.28867513459481281*y[67] + 0.16666666666666652*y[61] - 0.33333333333333326*y[62])/(p[23]*p[27]);
out[27] = -p[26]*(0.16666666666666677*y[66] - 0.28867513459481275*y[60] + 0.16666666666666652*y[67] + 0.28867513459481281*y[61] - 0.33333333333333326*y[68])/(p[23]*p[27]);
out[28] = -p[26]*(-0.33333333333333331*y[24] + 0.28867513459481287*y[27] + 0.16666666666666657*y[26] - 0.28867513459481275*y[29] + 0.16666666666666677*y[28])/(p[23]*p[27]);
out[29] = -p[26]*(-0.28867513459481287*y[25] + 0.16666666666666657*y[24] - 0.33333333333333326*y[26] + 0.28867513459481281*y[29] + 0.16666666666666652*y[28])/(p[23]*p[27]);
out[30] = -p[26]*(0.28867513459481275*y[25] + 0.16666666666666677*y[24] - 0.28867513459481281*y[27] + 0.16666666666666652*y[26] - 0.33333333333333326*y[28])/(p[23]*p[27]);
out[31] = -p[26]*(-0.33333333333333331*y[25] + 0.16666666666666657*y[27] - 0.28867513459481287*y[26] + 0.16666666666666677*y[29] + 0.28867513459481275*y[28])/(p[23]*p[27]);
out[32] = -p[26]*(0.16666666666666657*y[25] + 0.28867513459481287*y[24] - 0.33333333333333326*y[27] + 0.16666666666666652*y[29] - 0.28867513459481281*y[28])/(p[23]*p[27]);
out[33] = -p[26]*(0.16666666666666677*y[25] - 0.28867513459481275*y[24] + 0.16666666666666652*y[27] + 0.28867513459481281*y[26] - 0.33333333333333326*y[29])/(p[23]*p[27]);
out[223] = y[32];
out[224] = y[33];
out[225] = -y[32];
out[226] = -y[33];
out[227] = y[0] - y[6];
out[228] = y[1] - y[7];
out[229] = y[34];
out[230] = y[35];
out[231] = -y[34];
out[232] = -y[35];
out[233] = y[2] - y[6];
out[234] = y[3] - y[7];
out[235] = y[36];
out[236] = y[37];
out[237] = -y[36];
out[238] = -y[37];
out[239] = y[4] - y[6];
out[240] = y[5] - y[7];
out[241] = -y[33];
out[242] = y[32];
out[243] = y[33];
out[244] = -y[32];
out[245] = y[1] - y[7];
out[246] = -y[0] + y[6];
out[247] = -y[35];
out[248] = y[34];
out[249] = y[35];
out[250] = -y[34];
out[251] = y[3] - y[7];
out[252] = -y[2] + y[6];
out[253] = -y[37];
out[254] = y[36];
out[255] = y[37];
out[256] = -y[36];
out[257] = y[5] - y[7];
out[258] = -y[4] + y[6];
out[267] = y[40];
out[268] = y[41];
out[269] = -y[40];
out[270] = -y[41];
out[271] = y[8] - y[14];
out[272] = y[9] - y[15];
out[273] = y[42];
out[274] = y[43];
out[275] = -y[42];
out[276] = -y[43];
out[277] = y[10] - y[14];
out[278] = y[11] - y[15];
out[279] = y[44];
out[280] = y[45];
out[281] = -y[44];
out[282] = -y[45];
out[283] = y[12] - y[14];
out[284] = y[13] - y[15];
out[285] = -y[41];
out[286] = y[40];
out[287] = y[41];
out[288] = -y[40];
out[289] = y[9] - y[15];
out[290] = -y[8] + y[14];
out[291] = -y[43];
out[292] = y[42];
out[293] = y[43];
out[294] = -y[42];
out[295] = y[11] - y[15];
out[296] = -y[10] + y[14];
out[297] = -y[45];
out[298] = y[44];
out[299] = y[45];
out[300] = -y[44];
out[301] = y[13] - y[15];
out[302] = -y[12] + y[14];
out[311] = -u[16]*sin(x[0]);
out[316] = -u[17]*sin(x[0] - 2.0943951023931953);
out[321] = -u[18]*sin(x[0] - 4.1887902047863905);
out[338] = 1.0*u[16]*cos(x[0]);
out[343] = 1.0*u[17]*cos(x[0] - 2.0943951023931953);
out[348] = 1.0*u[18]*cos(x[0] - 4.1887902047863905);
out[365] = -u[22]*sin(x[2]);
out[370] = -u[23]*sin(x[2] - 2.0943951023931953);
out[375] = -u[24]*sin(x[2] - 4.1887902047863905);
out[392] = 1.0*u[22]*cos(x[2]);
out[397] = 1.0*u[23]*cos(x[2] - 2.0943951023931953);
out[402] = 1.0*u[24]*cos(x[2] - 4.1887902047863905);

}

void sp_jac_ini_up_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = -p[14];
out[3] = -1/p[12];
out[4] = -p[11]*p[13]/(p[8]*p[12]);
out[17] = -p[29];
out[20] = -1/p[27];
out[21] = -p[26]*p[28]/(p[23]*p[27]);
out[34] = -p[30];
out[314] = -p[1];
out[315] = 1.0*p[0];
out[319] = -p[3];
out[320] = 1.0*p[2];
out[324] = -p[5];
out[325] = 1.0*p[4];
out[327] = -p[7];
out[329] = 1.0*p[6];
out[335] = p[10];
out[337] = -1.0*p[9];
out[341] = -1.0*p[0];
out[342] = -1.0*p[1];
out[346] = -1.0*p[2];
out[347] = -1.0*p[3];
out[351] = -1.0*p[4];
out[352] = -1.0*p[5];
out[354] = -1.0*p[6];
out[355] = -1.0*p[7];
out[362] = 1.0*p[9];
out[363] = 1.0*p[10];
out[368] = -p[16];
out[369] = 1.0*p[15];
out[373] = -p[18];
out[374] = 1.0*p[17];
out[378] = -p[20];
out[379] = 1.0*p[19];
out[381] = -p[22];
out[383] = 1.0*p[21];
out[389] = p[25];
out[391] = -1.0*p[24];
out[395] = -1.0*p[15];
out[396] = -1.0*p[16];
out[400] = -1.0*p[17];
out[401] = -1.0*p[18];
out[405] = -1.0*p[19];
out[406] = -1.0*p[20];
out[408] = -1.0*p[21];
out[409] = -1.0*p[22];
out[416] = 1.0*p[24];
out[417] = 1.0*p[25];
out[419] = -p[8]/(p[8] + p[23]);
out[420] = -p[23]/(p[8] + p[23]);

}

void sp_jac_ini_num_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[1] = 314.15926535897933;
out[2] = -314.15926535897933;
out[18] = 314.15926535897933;
out[19] = -314.15926535897933;
out[35] = -292.22199539210823;
out[36] = -139.98658461897401;
out[37] = 48.703665898684704;
out[38] = 23.331097436495664;
out[39] = 243.51832949342352;
out[40] = 116.65548718247834;
out[41] = 1;
out[42] = 139.98658461897401;
out[43] = -292.22199539210823;
out[44] = -23.331097436495664;
out[45] = 48.703665898684704;
out[46] = -116.65548718247834;
out[47] = 243.51832949342352;
out[48] = 1;
out[49] = -292.22199539210823;
out[50] = -139.98658461897401;
out[51] = 48.703665898684704;
out[52] = 23.331097436495664;
out[53] = 243.51832949342352;
out[54] = 116.65548718247834;
out[55] = 1;
out[56] = 139.98658461897401;
out[57] = -292.22199539210823;
out[58] = -23.331097436495664;
out[59] = 48.703665898684704;
out[60] = -116.65548718247834;
out[61] = 243.51832949342352;
out[62] = 1;
out[63] = -292.22199539210823;
out[64] = -139.98658461897401;
out[65] = 48.703665898684704;
out[66] = 23.331097436495664;
out[67] = 243.51832949342352;
out[68] = 116.65548718247834;
out[69] = 1;
out[70] = 139.98658461897401;
out[71] = -292.22199539210823;
out[72] = -23.331097436495664;
out[73] = 48.703665898684704;
out[74] = -116.65548718247834;
out[75] = 243.51832949342352;
out[76] = 1;
out[77] = -292.22199539210823;
out[78] = -139.98658461897401;
out[79] = 48.703665898684704;
out[80] = 23.331097436495664;
out[81] = 243.51832949342352;
out[82] = 116.65548718247834;
out[83] = 139.98658461897401;
out[84] = -292.22199539210823;
out[85] = -23.331097436495664;
out[86] = 48.703665898684704;
out[87] = -116.65548718247834;
out[88] = 243.51832949342352;
out[89] = 48.703665898684704;
out[90] = 23.331097436495664;
out[91] = -292.22199539210823;
out[92] = -139.98658461897401;
out[93] = 243.51832949342352;
out[94] = 116.65548718247834;
out[95] = 1;
out[96] = -23.331097436495664;
out[97] = 48.703665898684704;
out[98] = 139.98658461897401;
out[99] = -292.22199539210823;
out[100] = -116.65548718247834;
out[101] = 243.51832949342352;
out[102] = 1;
out[103] = 48.703665898684704;
out[104] = 23.331097436495664;
out[105] = -292.22199539210823;
out[106] = -139.98658461897401;
out[107] = 243.51832949342352;
out[108] = 116.65548718247834;
out[109] = 1;
out[110] = -23.331097436495664;
out[111] = 48.703665898684704;
out[112] = 139.98658461897401;
out[113] = -292.22199539210823;
out[114] = -116.65548718247834;
out[115] = 243.51832949342352;
out[116] = 1;
out[117] = 48.703665898684704;
out[118] = 23.331097436495664;
out[119] = -292.22199539210823;
out[120] = -139.98658461897401;
out[121] = 243.51832949342352;
out[122] = 116.65548718247834;
out[123] = 1;
out[124] = -23.331097436495664;
out[125] = 48.703665898684704;
out[126] = 139.98658461897401;
out[127] = -292.22199539210823;
out[128] = -116.65548718247834;
out[129] = 243.51832949342352;
out[130] = 1;
out[131] = 48.703665898684704;
out[132] = 23.331097436495664;
out[133] = -292.22199539210823;
out[134] = -139.98658461897401;
out[135] = 243.51832949342352;
out[136] = 116.65548718247834;
out[137] = -23.331097436495664;
out[138] = 48.703665898684704;
out[139] = 139.98658461897401;
out[140] = -292.22199539210823;
out[141] = -116.65548718247834;
out[142] = 243.51832949342352;
out[143] = 243.51832949342352;
out[144] = 116.65548718247834;
out[145] = -243.51832949342352;
out[146] = -116.65548718247834;
out[147] = 1;
out[148] = -116.65548718247834;
out[149] = 243.51832949342352;
out[150] = 116.65548718247834;
out[151] = -243.51832949342352;
out[152] = 1;
out[153] = 243.51832949342352;
out[154] = 116.65548718247834;
out[155] = -243.51832949342352;
out[156] = -116.65548718247834;
out[157] = 1;
out[158] = -116.65548718247834;
out[159] = 243.51832949342352;
out[160] = 116.65548718247834;
out[161] = -243.51832949342352;
out[162] = 1;
out[163] = 243.51832949342352;
out[164] = 116.65548718247834;
out[165] = -243.51832949342352;
out[166] = -116.65548718247834;
out[167] = 1;
out[168] = -116.65548718247834;
out[169] = 243.51832949342352;
out[170] = 116.65548718247834;
out[171] = -243.51832949342352;
out[172] = 1;
out[173] = 243.51832949342352;
out[174] = 116.65548718247834;
out[175] = -243.51832949342352;
out[176] = -116.65548718247834;
out[177] = 1;
out[178] = -116.65548718247834;
out[179] = 243.51832949342352;
out[180] = 116.65548718247834;
out[181] = -243.51832949342352;
out[182] = 1;
out[183] = 243.51832949342352;
out[184] = 116.65548718247834;
out[185] = -243.51832949342352;
out[186] = -116.65548718247834;
out[187] = 1;
out[188] = -116.65548718247834;
out[189] = 243.51832949342352;
out[190] = 116.65548718247834;
out[191] = -243.51832949342352;
out[192] = 1;
out[193] = 243.51832949342352;
out[194] = 116.65548718247834;
out[195] = -243.51832949342352;
out[196] = -116.65548718247834;
out[197] = 1;
out[198] = -116.65548718247834;
out[199] = 243.51832949342352;
out[200] = 116.65548718247834;
out[201] = -243.51832949342352;
out[202] = 1;
out[203] = 243.51832949342352;
out[204] = 116.65548718247834;
out[205] = -243.51832949342352;
out[206] = -116.65548718247834;
out[207] = 1;
out[208] = -116.65548718247834;
out[209] = 243.51832949342352;
out[210] = 116.65548718247834;
out[211] = -243.51832949342352;
out[212] = 1;
out[213] = 243.51832949342352;
out[214] = 116.65548718247834;
out[215] = -243.51832949342352;
out[216] = -116.65548718247834;
out[217] = 1;
out[218] = -116.65548718247834;
out[219] = 243.51832949342352;
out[220] = 116.65548718247834;
out[221] = -243.51832949342352;
out[222] = 1;
out[259] = 1;
out[260] = 1;
out[261] = 1;
out[262] = 1;
out[263] = 1;
out[264] = 1;
out[265] = 1;
out[266] = 1;
out[303] = 1;
out[304] = 1;
out[305] = 1;
out[306] = 1;
out[307] = 1;
out[308] = 1;
out[309] = 1;
out[310] = 1;
out[312] = -1;
out[313] = 1;
out[317] = -1;
out[318] = 1;
out[322] = -1;
out[323] = 1;
out[326] = -1;
out[328] = 1;
out[330] = 1;
out[331] = 1;
out[332] = 1;
out[333] = 1;
out[334] = -1;
out[336] = -1;
out[339] = -1.0;
out[340] = 1.0;
out[344] = -1.0;
out[345] = 1.0;
out[349] = -1.0;
out[350] = 1.0;
out[353] = -1.0;
out[356] = 1.0;
out[357] = 1.0;
out[358] = 1.0;
out[359] = 1.0;
out[360] = 1.0;
out[361] = -1.0;
out[364] = -1.0;
out[366] = -1;
out[367] = 1;
out[371] = -1;
out[372] = 1;
out[376] = -1;
out[377] = 1;
out[380] = -1;
out[382] = 1;
out[384] = 1;
out[385] = 1;
out[386] = 1;
out[387] = 1;
out[388] = -1;
out[390] = -1;
out[393] = -1.0;
out[394] = 1.0;
out[398] = -1.0;
out[399] = 1.0;
out[403] = -1.0;
out[404] = 1.0;
out[407] = -1.0;
out[410] = 1.0;
out[411] = 1.0;
out[412] = 1.0;
out[413] = 1.0;
out[414] = 1.0;
out[415] = -1.0;
out[418] = -1.0;
out[421] = 1;

}

void de_jac_trap_xy_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[99] = 0.5*Dt*p[11]*(-0.33333333333333331*y[48] + 0.28867513459481287*y[55] + 0.16666666666666657*y[49] - 0.28867513459481275*y[56] + 0.16666666666666677*y[50])/(p[8]*p[12]);
out[100] = 0.5*Dt*p[11]*(-0.33333333333333331*y[54] + 0.16666666666666657*y[55] - 0.28867513459481287*y[49] + 0.16666666666666677*y[56] + 0.28867513459481275*y[50])/(p[8]*p[12]);
out[101] = 0.5*Dt*p[11]*(-0.28867513459481287*y[54] + 0.16666666666666657*y[48] - 0.33333333333333326*y[49] + 0.28867513459481281*y[56] + 0.16666666666666652*y[50])/(p[8]*p[12]);
out[102] = 0.5*Dt*p[11]*(0.16666666666666657*y[54] + 0.28867513459481287*y[48] - 0.33333333333333326*y[55] + 0.16666666666666652*y[56] - 0.28867513459481281*y[50])/(p[8]*p[12]);
out[103] = 0.5*Dt*p[11]*(0.28867513459481275*y[54] + 0.16666666666666677*y[48] - 0.28867513459481281*y[55] + 0.16666666666666652*y[49] - 0.33333333333333326*y[50])/(p[8]*p[12]);
out[104] = 0.5*Dt*p[11]*(0.16666666666666677*y[54] - 0.28867513459481275*y[48] + 0.16666666666666652*y[55] + 0.28867513459481281*y[49] - 0.33333333333333326*y[56])/(p[8]*p[12]);
out[131] = 0.5*Dt*p[11]*(-0.33333333333333331*y[16] + 0.28867513459481287*y[19] + 0.16666666666666657*y[18] - 0.28867513459481275*y[21] + 0.16666666666666677*y[20])/(p[8]*p[12]);
out[132] = 0.5*Dt*p[11]*(-0.28867513459481287*y[17] + 0.16666666666666657*y[16] - 0.33333333333333326*y[18] + 0.28867513459481281*y[21] + 0.16666666666666652*y[20])/(p[8]*p[12]);
out[133] = 0.5*Dt*p[11]*(0.28867513459481275*y[17] + 0.16666666666666677*y[16] - 0.28867513459481281*y[19] + 0.16666666666666652*y[18] - 0.33333333333333326*y[20])/(p[8]*p[12]);
out[137] = 0.5*Dt*p[11]*(-0.33333333333333331*y[17] + 0.16666666666666657*y[19] - 0.28867513459481287*y[18] + 0.16666666666666677*y[21] + 0.28867513459481275*y[20])/(p[8]*p[12]);
out[138] = 0.5*Dt*p[11]*(0.16666666666666657*y[17] + 0.28867513459481287*y[16] - 0.33333333333333326*y[19] + 0.16666666666666652*y[21] - 0.28867513459481281*y[20])/(p[8]*p[12]);
out[139] = 0.5*Dt*p[11]*(0.16666666666666677*y[17] - 0.28867513459481275*y[16] + 0.16666666666666652*y[19] + 0.28867513459481281*y[18] - 0.33333333333333326*y[21])/(p[8]*p[12]);
out[263] = 0.5*Dt*p[26]*(-0.33333333333333331*y[60] + 0.28867513459481287*y[67] + 0.16666666666666657*y[61] - 0.28867513459481275*y[68] + 0.16666666666666677*y[62])/(p[23]*p[27]);
out[264] = 0.5*Dt*p[26]*(-0.33333333333333331*y[66] + 0.16666666666666657*y[67] - 0.28867513459481287*y[61] + 0.16666666666666677*y[68] + 0.28867513459481275*y[62])/(p[23]*p[27]);
out[265] = 0.5*Dt*p[26]*(-0.28867513459481287*y[66] + 0.16666666666666657*y[60] - 0.33333333333333326*y[61] + 0.28867513459481281*y[68] + 0.16666666666666652*y[62])/(p[23]*p[27]);
out[266] = 0.5*Dt*p[26]*(0.16666666666666657*y[66] + 0.28867513459481287*y[60] - 0.33333333333333326*y[67] + 0.16666666666666652*y[68] - 0.28867513459481281*y[62])/(p[23]*p[27]);
out[267] = 0.5*Dt*p[26]*(0.28867513459481275*y[66] + 0.16666666666666677*y[60] - 0.28867513459481281*y[67] + 0.16666666666666652*y[61] - 0.33333333333333326*y[62])/(p[23]*p[27]);
out[268] = 0.5*Dt*p[26]*(0.16666666666666677*y[66] - 0.28867513459481275*y[60] + 0.16666666666666652*y[67] + 0.28867513459481281*y[61] - 0.33333333333333326*y[68])/(p[23]*p[27]);
out[299] = 0.5*Dt*p[26]*(-0.33333333333333331*y[24] + 0.28867513459481287*y[27] + 0.16666666666666657*y[26] - 0.28867513459481275*y[29] + 0.16666666666666677*y[28])/(p[23]*p[27]);
out[300] = 0.5*Dt*p[26]*(-0.28867513459481287*y[25] + 0.16666666666666657*y[24] - 0.33333333333333326*y[26] + 0.28867513459481281*y[29] + 0.16666666666666652*y[28])/(p[23]*p[27]);
out[301] = 0.5*Dt*p[26]*(0.28867513459481275*y[25] + 0.16666666666666677*y[24] - 0.28867513459481281*y[27] + 0.16666666666666652*y[26] - 0.33333333333333326*y[28])/(p[23]*p[27]);
out[305] = 0.5*Dt*p[26]*(-0.33333333333333331*y[25] + 0.16666666666666657*y[27] - 0.28867513459481287*y[26] + 0.16666666666666677*y[29] + 0.28867513459481275*y[28])/(p[23]*p[27]);
out[306] = 0.5*Dt*p[26]*(0.16666666666666657*y[25] + 0.28867513459481287*y[24] - 0.33333333333333326*y[27] + 0.16666666666666652*y[29] - 0.28867513459481281*y[28])/(p[23]*p[27]);
out[307] = 0.5*Dt*p[26]*(0.16666666666666677*y[25] - 0.28867513459481275*y[24] + 0.16666666666666652*y[27] + 0.28867513459481281*y[26] - 0.33333333333333326*y[29])/(p[23]*p[27]);
out[2891] = y[32];
out[2892] = y[33];
out[2897] = -y[32];
out[2898] = -y[33];
out[2923] = y[0] - y[6];
out[2924] = y[1] - y[7];
out[2971] = y[34];
out[2972] = y[35];
out[2975] = -y[34];
out[2976] = -y[35];
out[3003] = y[2] - y[6];
out[3004] = y[3] - y[7];
out[3051] = y[36];
out[3052] = y[37];
out[3053] = -y[36];
out[3054] = -y[37];
out[3083] = y[4] - y[6];
out[3084] = y[5] - y[7];
out[3125] = -y[33];
out[3126] = y[32];
out[3131] = y[33];
out[3132] = -y[32];
out[3157] = y[1] - y[7];
out[3158] = -y[0] + y[6];
out[3205] = -y[35];
out[3206] = y[34];
out[3209] = y[35];
out[3210] = -y[34];
out[3237] = y[3] - y[7];
out[3238] = -y[2] + y[6];
out[3285] = -y[37];
out[3286] = y[36];
out[3287] = y[37];
out[3288] = -y[36];
out[3317] = y[5] - y[7];
out[3318] = -y[4] + y[6];
out[3523] = y[40];
out[3524] = y[41];
out[3529] = -y[40];
out[3530] = -y[41];
out[3555] = y[8] - y[14];
out[3556] = y[9] - y[15];
out[3603] = y[42];
out[3604] = y[43];
out[3607] = -y[42];
out[3608] = -y[43];
out[3635] = y[10] - y[14];
out[3636] = y[11] - y[15];
out[3683] = y[44];
out[3684] = y[45];
out[3685] = -y[44];
out[3686] = -y[45];
out[3715] = y[12] - y[14];
out[3716] = y[13] - y[15];
out[3757] = -y[41];
out[3758] = y[40];
out[3763] = y[41];
out[3764] = -y[40];
out[3789] = y[9] - y[15];
out[3790] = -y[8] + y[14];
out[3837] = -y[43];
out[3838] = y[42];
out[3841] = y[43];
out[3842] = -y[42];
out[3869] = y[11] - y[15];
out[3870] = -y[10] + y[14];
out[3917] = -y[45];
out[3918] = y[44];
out[3919] = y[45];
out[3920] = -y[44];
out[3949] = y[13] - y[15];
out[3950] = -y[12] + y[14];
out[4134] = -u[16]*sin(x[0]);
out[4212] = -u[17]*sin(x[0] - 2.0943951023931953);
out[4290] = -u[18]*sin(x[0] - 4.1887902047863905);
out[4602] = 1.0*u[16]*cos(x[0]);
out[4680] = 1.0*u[17]*cos(x[0] - 2.0943951023931953);
out[4758] = 1.0*u[18]*cos(x[0] - 4.1887902047863905);
out[5072] = -u[22]*sin(x[2]);
out[5150] = -u[23]*sin(x[2] - 2.0943951023931953);
out[5228] = -u[24]*sin(x[2] - 4.1887902047863905);
out[5540] = 1.0*u[22]*cos(x[2]);
out[5618] = 1.0*u[23]*cos(x[2] - 2.0943951023931953);
out[5696] = 1.0*u[24]*cos(x[2] - 4.1887902047863905);

}

void de_jac_trap_up_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = 0.5*Dt*p[14] + 1;
out[1] = -157.07963267948966*Dt;
out[77] = 157.07963267948966*Dt;
out[79] = 0.5*Dt/p[12] + 1;
out[82] = 0.5*Dt*p[11]*p[13]/(p[8]*p[12]);
out[158] = 0.5*Dt*p[29] + 1;
out[159] = -157.07963267948966*Dt;
out[233] = 157.07963267948966*Dt;
out[237] = 0.5*Dt/p[27] + 1;
out[238] = 0.5*Dt*p[26]*p[28]/(p[23]*p[27]);
out[389] = 0.5*Dt*p[30];
out[4187] = -p[1];
out[4193] = 1.0*p[0];
out[4266] = -p[3];
out[4272] = 1.0*p[2];
out[4345] = -p[5];
out[4351] = 1.0*p[4];
out[4424] = -p[7];
out[4430] = 1.0*p[6];
out[4581] = p[10];
out[4587] = -1.0*p[9];
out[4655] = -1.0*p[0];
out[4661] = -1.0*p[1];
out[4734] = -1.0*p[2];
out[4740] = -1.0*p[3];
out[4813] = -1.0*p[4];
out[4819] = -1.0*p[5];
out[4892] = -1.0*p[6];
out[4898] = -1.0*p[7];
out[5049] = 1.0*p[9];
out[5055] = 1.0*p[10];
out[5135] = -p[16];
out[5141] = 1.0*p[15];
out[5214] = -p[18];
out[5220] = 1.0*p[17];
out[5293] = -p[20];
out[5299] = 1.0*p[19];
out[5372] = -p[22];
out[5378] = 1.0*p[21];
out[5529] = p[25];
out[5535] = -1.0*p[24];
out[5603] = -1.0*p[15];
out[5609] = -1.0*p[16];
out[5682] = -1.0*p[17];
out[5688] = -1.0*p[18];
out[5761] = -1.0*p[19];
out[5767] = -1.0*p[20];
out[5840] = -1.0*p[21];
out[5846] = -1.0*p[22];
out[5997] = 1.0*p[24];
out[6003] = 1.0*p[25];
out[6007] = -p[8]/(p[8] + p[23]);
out[6009] = -p[23]/(p[8] + p[23]);

}

void de_jac_trap_num_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[316] = 1;
out[395] = -292.22199539210823;
out[396] = -139.98658461897401;
out[403] = 48.703665898684704;
out[404] = 23.331097436495664;
out[411] = 243.51832949342352;
out[412] = 116.65548718247834;
out[427] = 1;
out[473] = 139.98658461897401;
out[474] = -292.22199539210823;
out[481] = -23.331097436495664;
out[482] = 48.703665898684704;
out[489] = -116.65548718247834;
out[490] = 243.51832949342352;
out[506] = 1;
out[553] = -292.22199539210823;
out[554] = -139.98658461897401;
out[561] = 48.703665898684704;
out[562] = 23.331097436495664;
out[569] = 243.51832949342352;
out[570] = 116.65548718247834;
out[585] = 1;
out[631] = 139.98658461897401;
out[632] = -292.22199539210823;
out[639] = -23.331097436495664;
out[640] = 48.703665898684704;
out[647] = -116.65548718247834;
out[648] = 243.51832949342352;
out[664] = 1;
out[711] = -292.22199539210823;
out[712] = -139.98658461897401;
out[719] = 48.703665898684704;
out[720] = 23.331097436495664;
out[727] = 243.51832949342352;
out[728] = 116.65548718247834;
out[743] = 1;
out[789] = 139.98658461897401;
out[790] = -292.22199539210823;
out[797] = -23.331097436495664;
out[798] = 48.703665898684704;
out[805] = -116.65548718247834;
out[806] = 243.51832949342352;
out[822] = 1;
out[869] = -292.22199539210823;
out[870] = -139.98658461897401;
out[877] = 48.703665898684704;
out[878] = 23.331097436495664;
out[885] = 243.51832949342352;
out[886] = 116.65548718247834;
out[947] = 139.98658461897401;
out[948] = -292.22199539210823;
out[955] = -23.331097436495664;
out[956] = 48.703665898684704;
out[963] = -116.65548718247834;
out[964] = 243.51832949342352;
out[1019] = 48.703665898684704;
out[1020] = 23.331097436495664;
out[1027] = -292.22199539210823;
out[1028] = -139.98658461897401;
out[1043] = 243.51832949342352;
out[1044] = 116.65548718247834;
out[1059] = 1;
out[1097] = -23.331097436495664;
out[1098] = 48.703665898684704;
out[1105] = 139.98658461897401;
out[1106] = -292.22199539210823;
out[1121] = -116.65548718247834;
out[1122] = 243.51832949342352;
out[1138] = 1;
out[1177] = 48.703665898684704;
out[1178] = 23.331097436495664;
out[1185] = -292.22199539210823;
out[1186] = -139.98658461897401;
out[1201] = 243.51832949342352;
out[1202] = 116.65548718247834;
out[1217] = 1;
out[1255] = -23.331097436495664;
out[1256] = 48.703665898684704;
out[1263] = 139.98658461897401;
out[1264] = -292.22199539210823;
out[1279] = -116.65548718247834;
out[1280] = 243.51832949342352;
out[1296] = 1;
out[1335] = 48.703665898684704;
out[1336] = 23.331097436495664;
out[1343] = -292.22199539210823;
out[1344] = -139.98658461897401;
out[1359] = 243.51832949342352;
out[1360] = 116.65548718247834;
out[1375] = 1;
out[1413] = -23.331097436495664;
out[1414] = 48.703665898684704;
out[1421] = 139.98658461897401;
out[1422] = -292.22199539210823;
out[1437] = -116.65548718247834;
out[1438] = 243.51832949342352;
out[1454] = 1;
out[1493] = 48.703665898684704;
out[1494] = 23.331097436495664;
out[1501] = -292.22199539210823;
out[1502] = -139.98658461897401;
out[1517] = 243.51832949342352;
out[1518] = 116.65548718247834;
out[1571] = -23.331097436495664;
out[1572] = 48.703665898684704;
out[1579] = 139.98658461897401;
out[1580] = -292.22199539210823;
out[1595] = -116.65548718247834;
out[1596] = 243.51832949342352;
out[1643] = 243.51832949342352;
out[1644] = 116.65548718247834;
out[1659] = -243.51832949342352;
out[1660] = -116.65548718247834;
out[1691] = 1;
out[1721] = -116.65548718247834;
out[1722] = 243.51832949342352;
out[1737] = 116.65548718247834;
out[1738] = -243.51832949342352;
out[1775] = 1;
out[1801] = 243.51832949342352;
out[1802] = 116.65548718247834;
out[1817] = -243.51832949342352;
out[1818] = -116.65548718247834;
out[1848] = 1;
out[1879] = -116.65548718247834;
out[1880] = 243.51832949342352;
out[1895] = 116.65548718247834;
out[1896] = -243.51832949342352;
out[1932] = 1;
out[1959] = 243.51832949342352;
out[1960] = 116.65548718247834;
out[1975] = -243.51832949342352;
out[1976] = -116.65548718247834;
out[2005] = 1;
out[2037] = -116.65548718247834;
out[2038] = 243.51832949342352;
out[2053] = 116.65548718247834;
out[2054] = -243.51832949342352;
out[2089] = 1;
out[2117] = 243.51832949342352;
out[2118] = 116.65548718247834;
out[2133] = -243.51832949342352;
out[2134] = -116.65548718247834;
out[2162] = 1;
out[2195] = -116.65548718247834;
out[2196] = 243.51832949342352;
out[2211] = 116.65548718247834;
out[2212] = -243.51832949342352;
out[2246] = 1;
out[2275] = 243.51832949342352;
out[2276] = 116.65548718247834;
out[2291] = -243.51832949342352;
out[2292] = -116.65548718247834;
out[2327] = 1;
out[2353] = -116.65548718247834;
out[2354] = 243.51832949342352;
out[2369] = 116.65548718247834;
out[2370] = -243.51832949342352;
out[2411] = 1;
out[2433] = 243.51832949342352;
out[2434] = 116.65548718247834;
out[2449] = -243.51832949342352;
out[2450] = -116.65548718247834;
out[2484] = 1;
out[2511] = -116.65548718247834;
out[2512] = 243.51832949342352;
out[2527] = 116.65548718247834;
out[2528] = -243.51832949342352;
out[2568] = 1;
out[2591] = 243.51832949342352;
out[2592] = 116.65548718247834;
out[2607] = -243.51832949342352;
out[2608] = -116.65548718247834;
out[2641] = 1;
out[2669] = -116.65548718247834;
out[2670] = 243.51832949342352;
out[2685] = 116.65548718247834;
out[2686] = -243.51832949342352;
out[2725] = 1;
out[2749] = 243.51832949342352;
out[2750] = 116.65548718247834;
out[2765] = -243.51832949342352;
out[2766] = -116.65548718247834;
out[2798] = 1;
out[2827] = -116.65548718247834;
out[2828] = 243.51832949342352;
out[2843] = 116.65548718247834;
out[2844] = -243.51832949342352;
out[2882] = 1;
out[3391] = 1;
out[3393] = 1;
out[3395] = 1;
out[3397] = 1;
out[3470] = 1;
out[3472] = 1;
out[3474] = 1;
out[3476] = 1;
out[4023] = 1;
out[4025] = 1;
out[4027] = 1;
out[4029] = 1;
out[4102] = 1;
out[4104] = 1;
out[4106] = 1;
out[4108] = 1;
out[4155] = -1;
out[4161] = 1;
out[4235] = -1;
out[4239] = 1;
out[4315] = -1;
out[4317] = 1;
out[4395] = -1;
out[4426] = 1;
out[4499] = 1;
out[4500] = 1;
out[4501] = 1;
out[4502] = 1;
out[4503] = -1;
out[4582] = -1;
out[4624] = -1.0;
out[4630] = 1.0;
out[4704] = -1.0;
out[4708] = 1.0;
out[4784] = -1.0;
out[4786] = 1.0;
out[4864] = -1.0;
out[4900] = 1.0;
out[4973] = 1.0;
out[4974] = 1.0;
out[4975] = 1.0;
out[4976] = 1.0;
out[4977] = -1.0;
out[5056] = -1.0;
out[5099] = -1;
out[5105] = 1;
out[5179] = -1;
out[5183] = 1;
out[5259] = -1;
out[5261] = 1;
out[5339] = -1;
out[5374] = 1;
out[5447] = 1;
out[5448] = 1;
out[5449] = 1;
out[5450] = 1;
out[5451] = -1;
out[5530] = -1;
out[5568] = -1.0;
out[5574] = 1.0;
out[5648] = -1.0;
out[5652] = 1.0;
out[5728] = -1.0;
out[5730] = 1.0;
out[5808] = -1.0;
out[5848] = 1.0;
out[5921] = 1.0;
out[5922] = 1.0;
out[5923] = 1.0;
out[5924] = 1.0;
out[5925] = -1.0;
out[6004] = -1.0;
out[6083] = 1;

}

void sp_jac_trap_xy_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[5] = 0.5*Dt*p[11]*(-0.33333333333333331*y[48] + 0.28867513459481287*y[55] + 0.16666666666666657*y[49] - 0.28867513459481275*y[56] + 0.16666666666666677*y[50])/(p[8]*p[12]);
out[6] = 0.5*Dt*p[11]*(-0.33333333333333331*y[54] + 0.16666666666666657*y[55] - 0.28867513459481287*y[49] + 0.16666666666666677*y[56] + 0.28867513459481275*y[50])/(p[8]*p[12]);
out[7] = 0.5*Dt*p[11]*(-0.28867513459481287*y[54] + 0.16666666666666657*y[48] - 0.33333333333333326*y[49] + 0.28867513459481281*y[56] + 0.16666666666666652*y[50])/(p[8]*p[12]);
out[8] = 0.5*Dt*p[11]*(0.16666666666666657*y[54] + 0.28867513459481287*y[48] - 0.33333333333333326*y[55] + 0.16666666666666652*y[56] - 0.28867513459481281*y[50])/(p[8]*p[12]);
out[9] = 0.5*Dt*p[11]*(0.28867513459481275*y[54] + 0.16666666666666677*y[48] - 0.28867513459481281*y[55] + 0.16666666666666652*y[49] - 0.33333333333333326*y[50])/(p[8]*p[12]);
out[10] = 0.5*Dt*p[11]*(0.16666666666666677*y[54] - 0.28867513459481275*y[48] + 0.16666666666666652*y[55] + 0.28867513459481281*y[49] - 0.33333333333333326*y[56])/(p[8]*p[12]);
out[11] = 0.5*Dt*p[11]*(-0.33333333333333331*y[16] + 0.28867513459481287*y[19] + 0.16666666666666657*y[18] - 0.28867513459481275*y[21] + 0.16666666666666677*y[20])/(p[8]*p[12]);
out[12] = 0.5*Dt*p[11]*(-0.28867513459481287*y[17] + 0.16666666666666657*y[16] - 0.33333333333333326*y[18] + 0.28867513459481281*y[21] + 0.16666666666666652*y[20])/(p[8]*p[12]);
out[13] = 0.5*Dt*p[11]*(0.28867513459481275*y[17] + 0.16666666666666677*y[16] - 0.28867513459481281*y[19] + 0.16666666666666652*y[18] - 0.33333333333333326*y[20])/(p[8]*p[12]);
out[14] = 0.5*Dt*p[11]*(-0.33333333333333331*y[17] + 0.16666666666666657*y[19] - 0.28867513459481287*y[18] + 0.16666666666666677*y[21] + 0.28867513459481275*y[20])/(p[8]*p[12]);
out[15] = 0.5*Dt*p[11]*(0.16666666666666657*y[17] + 0.28867513459481287*y[16] - 0.33333333333333326*y[19] + 0.16666666666666652*y[21] - 0.28867513459481281*y[20])/(p[8]*p[12]);
out[16] = 0.5*Dt*p[11]*(0.16666666666666677*y[17] - 0.28867513459481275*y[16] + 0.16666666666666652*y[19] + 0.28867513459481281*y[18] - 0.33333333333333326*y[21])/(p[8]*p[12]);
out[22] = 0.5*Dt*p[26]*(-0.33333333333333331*y[60] + 0.28867513459481287*y[67] + 0.16666666666666657*y[61] - 0.28867513459481275*y[68] + 0.16666666666666677*y[62])/(p[23]*p[27]);
out[23] = 0.5*Dt*p[26]*(-0.33333333333333331*y[66] + 0.16666666666666657*y[67] - 0.28867513459481287*y[61] + 0.16666666666666677*y[68] + 0.28867513459481275*y[62])/(p[23]*p[27]);
out[24] = 0.5*Dt*p[26]*(-0.28867513459481287*y[66] + 0.16666666666666657*y[60] - 0.33333333333333326*y[61] + 0.28867513459481281*y[68] + 0.16666666666666652*y[62])/(p[23]*p[27]);
out[25] = 0.5*Dt*p[26]*(0.16666666666666657*y[66] + 0.28867513459481287*y[60] - 0.33333333333333326*y[67] + 0.16666666666666652*y[68] - 0.28867513459481281*y[62])/(p[23]*p[27]);
out[26] = 0.5*Dt*p[26]*(0.28867513459481275*y[66] + 0.16666666666666677*y[60] - 0.28867513459481281*y[67] + 0.16666666666666652*y[61] - 0.33333333333333326*y[62])/(p[23]*p[27]);
out[27] = 0.5*Dt*p[26]*(0.16666666666666677*y[66] - 0.28867513459481275*y[60] + 0.16666666666666652*y[67] + 0.28867513459481281*y[61] - 0.33333333333333326*y[68])/(p[23]*p[27]);
out[28] = 0.5*Dt*p[26]*(-0.33333333333333331*y[24] + 0.28867513459481287*y[27] + 0.16666666666666657*y[26] - 0.28867513459481275*y[29] + 0.16666666666666677*y[28])/(p[23]*p[27]);
out[29] = 0.5*Dt*p[26]*(-0.28867513459481287*y[25] + 0.16666666666666657*y[24] - 0.33333333333333326*y[26] + 0.28867513459481281*y[29] + 0.16666666666666652*y[28])/(p[23]*p[27]);
out[30] = 0.5*Dt*p[26]*(0.28867513459481275*y[25] + 0.16666666666666677*y[24] - 0.28867513459481281*y[27] + 0.16666666666666652*y[26] - 0.33333333333333326*y[28])/(p[23]*p[27]);
out[31] = 0.5*Dt*p[26]*(-0.33333333333333331*y[25] + 0.16666666666666657*y[27] - 0.28867513459481287*y[26] + 0.16666666666666677*y[29] + 0.28867513459481275*y[28])/(p[23]*p[27]);
out[32] = 0.5*Dt*p[26]*(0.16666666666666657*y[25] + 0.28867513459481287*y[24] - 0.33333333333333326*y[27] + 0.16666666666666652*y[29] - 0.28867513459481281*y[28])/(p[23]*p[27]);
out[33] = 0.5*Dt*p[26]*(0.16666666666666677*y[25] - 0.28867513459481275*y[24] + 0.16666666666666652*y[27] + 0.28867513459481281*y[26] - 0.33333333333333326*y[29])/(p[23]*p[27]);
out[224] = y[32];
out[225] = y[33];
out[226] = -y[32];
out[227] = -y[33];
out[228] = y[0] - y[6];
out[229] = y[1] - y[7];
out[230] = y[34];
out[231] = y[35];
out[232] = -y[34];
out[233] = -y[35];
out[234] = y[2] - y[6];
out[235] = y[3] - y[7];
out[236] = y[36];
out[237] = y[37];
out[238] = -y[36];
out[239] = -y[37];
out[240] = y[4] - y[6];
out[241] = y[5] - y[7];
out[242] = -y[33];
out[243] = y[32];
out[244] = y[33];
out[245] = -y[32];
out[246] = y[1] - y[7];
out[247] = -y[0] + y[6];
out[248] = -y[35];
out[249] = y[34];
out[250] = y[35];
out[251] = -y[34];
out[252] = y[3] - y[7];
out[253] = -y[2] + y[6];
out[254] = -y[37];
out[255] = y[36];
out[256] = y[37];
out[257] = -y[36];
out[258] = y[5] - y[7];
out[259] = -y[4] + y[6];
out[268] = y[40];
out[269] = y[41];
out[270] = -y[40];
out[271] = -y[41];
out[272] = y[8] - y[14];
out[273] = y[9] - y[15];
out[274] = y[42];
out[275] = y[43];
out[276] = -y[42];
out[277] = -y[43];
out[278] = y[10] - y[14];
out[279] = y[11] - y[15];
out[280] = y[44];
out[281] = y[45];
out[282] = -y[44];
out[283] = -y[45];
out[284] = y[12] - y[14];
out[285] = y[13] - y[15];
out[286] = -y[41];
out[287] = y[40];
out[288] = y[41];
out[289] = -y[40];
out[290] = y[9] - y[15];
out[291] = -y[8] + y[14];
out[292] = -y[43];
out[293] = y[42];
out[294] = y[43];
out[295] = -y[42];
out[296] = y[11] - y[15];
out[297] = -y[10] + y[14];
out[298] = -y[45];
out[299] = y[44];
out[300] = y[45];
out[301] = -y[44];
out[302] = y[13] - y[15];
out[303] = -y[12] + y[14];
out[312] = -u[16]*sin(x[0]);
out[317] = -u[17]*sin(x[0] - 2.0943951023931953);
out[322] = -u[18]*sin(x[0] - 4.1887902047863905);
out[339] = 1.0*u[16]*cos(x[0]);
out[344] = 1.0*u[17]*cos(x[0] - 2.0943951023931953);
out[349] = 1.0*u[18]*cos(x[0] - 4.1887902047863905);
out[366] = -u[22]*sin(x[2]);
out[371] = -u[23]*sin(x[2] - 2.0943951023931953);
out[376] = -u[24]*sin(x[2] - 4.1887902047863905);
out[393] = 1.0*u[22]*cos(x[2]);
out[398] = 1.0*u[23]*cos(x[2] - 2.0943951023931953);
out[403] = 1.0*u[24]*cos(x[2] - 4.1887902047863905);

}

void sp_jac_trap_up_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[0] = 0.5*Dt*p[14] + 1;
out[1] = -157.07963267948966*Dt;
out[2] = 157.07963267948966*Dt;
out[3] = 0.5*Dt/p[12] + 1;
out[4] = 0.5*Dt*p[11]*p[13]/(p[8]*p[12]);
out[17] = 0.5*Dt*p[29] + 1;
out[18] = -157.07963267948966*Dt;
out[19] = 157.07963267948966*Dt;
out[20] = 0.5*Dt/p[27] + 1;
out[21] = 0.5*Dt*p[26]*p[28]/(p[23]*p[27]);
out[35] = 0.5*Dt*p[30];
out[315] = -p[1];
out[316] = 1.0*p[0];
out[320] = -p[3];
out[321] = 1.0*p[2];
out[325] = -p[5];
out[326] = 1.0*p[4];
out[328] = -p[7];
out[330] = 1.0*p[6];
out[336] = p[10];
out[338] = -1.0*p[9];
out[342] = -1.0*p[0];
out[343] = -1.0*p[1];
out[347] = -1.0*p[2];
out[348] = -1.0*p[3];
out[352] = -1.0*p[4];
out[353] = -1.0*p[5];
out[355] = -1.0*p[6];
out[356] = -1.0*p[7];
out[363] = 1.0*p[9];
out[364] = 1.0*p[10];
out[369] = -p[16];
out[370] = 1.0*p[15];
out[374] = -p[18];
out[375] = 1.0*p[17];
out[379] = -p[20];
out[380] = 1.0*p[19];
out[382] = -p[22];
out[384] = 1.0*p[21];
out[390] = p[25];
out[392] = -1.0*p[24];
out[396] = -1.0*p[15];
out[397] = -1.0*p[16];
out[401] = -1.0*p[17];
out[402] = -1.0*p[18];
out[406] = -1.0*p[19];
out[407] = -1.0*p[20];
out[409] = -1.0*p[21];
out[410] = -1.0*p[22];
out[417] = 1.0*p[24];
out[418] = 1.0*p[25];
out[420] = -p[8]/(p[8] + p[23]);
out[421] = -p[23]/(p[8] + p[23]);

}

void sp_jac_trap_num_eval(double *out,double *x,double *y,double *u,double *p,double Dt){

out[34] = 1;
out[36] = -292.22199539210823;
out[37] = -139.98658461897401;
out[38] = 48.703665898684704;
out[39] = 23.331097436495664;
out[40] = 243.51832949342352;
out[41] = 116.65548718247834;
out[42] = 1;
out[43] = 139.98658461897401;
out[44] = -292.22199539210823;
out[45] = -23.331097436495664;
out[46] = 48.703665898684704;
out[47] = -116.65548718247834;
out[48] = 243.51832949342352;
out[49] = 1;
out[50] = -292.22199539210823;
out[51] = -139.98658461897401;
out[52] = 48.703665898684704;
out[53] = 23.331097436495664;
out[54] = 243.51832949342352;
out[55] = 116.65548718247834;
out[56] = 1;
out[57] = 139.98658461897401;
out[58] = -292.22199539210823;
out[59] = -23.331097436495664;
out[60] = 48.703665898684704;
out[61] = -116.65548718247834;
out[62] = 243.51832949342352;
out[63] = 1;
out[64] = -292.22199539210823;
out[65] = -139.98658461897401;
out[66] = 48.703665898684704;
out[67] = 23.331097436495664;
out[68] = 243.51832949342352;
out[69] = 116.65548718247834;
out[70] = 1;
out[71] = 139.98658461897401;
out[72] = -292.22199539210823;
out[73] = -23.331097436495664;
out[74] = 48.703665898684704;
out[75] = -116.65548718247834;
out[76] = 243.51832949342352;
out[77] = 1;
out[78] = -292.22199539210823;
out[79] = -139.98658461897401;
out[80] = 48.703665898684704;
out[81] = 23.331097436495664;
out[82] = 243.51832949342352;
out[83] = 116.65548718247834;
out[84] = 139.98658461897401;
out[85] = -292.22199539210823;
out[86] = -23.331097436495664;
out[87] = 48.703665898684704;
out[88] = -116.65548718247834;
out[89] = 243.51832949342352;
out[90] = 48.703665898684704;
out[91] = 23.331097436495664;
out[92] = -292.22199539210823;
out[93] = -139.98658461897401;
out[94] = 243.51832949342352;
out[95] = 116.65548718247834;
out[96] = 1;
out[97] = -23.331097436495664;
out[98] = 48.703665898684704;
out[99] = 139.98658461897401;
out[100] = -292.22199539210823;
out[101] = -116.65548718247834;
out[102] = 243.51832949342352;
out[103] = 1;
out[104] = 48.703665898684704;
out[105] = 23.331097436495664;
out[106] = -292.22199539210823;
out[107] = -139.98658461897401;
out[108] = 243.51832949342352;
out[109] = 116.65548718247834;
out[110] = 1;
out[111] = -23.331097436495664;
out[112] = 48.703665898684704;
out[113] = 139.98658461897401;
out[114] = -292.22199539210823;
out[115] = -116.65548718247834;
out[116] = 243.51832949342352;
out[117] = 1;
out[118] = 48.703665898684704;
out[119] = 23.331097436495664;
out[120] = -292.22199539210823;
out[121] = -139.98658461897401;
out[122] = 243.51832949342352;
out[123] = 116.65548718247834;
out[124] = 1;
out[125] = -23.331097436495664;
out[126] = 48.703665898684704;
out[127] = 139.98658461897401;
out[128] = -292.22199539210823;
out[129] = -116.65548718247834;
out[130] = 243.51832949342352;
out[131] = 1;
out[132] = 48.703665898684704;
out[133] = 23.331097436495664;
out[134] = -292.22199539210823;
out[135] = -139.98658461897401;
out[136] = 243.51832949342352;
out[137] = 116.65548718247834;
out[138] = -23.331097436495664;
out[139] = 48.703665898684704;
out[140] = 139.98658461897401;
out[141] = -292.22199539210823;
out[142] = -116.65548718247834;
out[143] = 243.51832949342352;
out[144] = 243.51832949342352;
out[145] = 116.65548718247834;
out[146] = -243.51832949342352;
out[147] = -116.65548718247834;
out[148] = 1;
out[149] = -116.65548718247834;
out[150] = 243.51832949342352;
out[151] = 116.65548718247834;
out[152] = -243.51832949342352;
out[153] = 1;
out[154] = 243.51832949342352;
out[155] = 116.65548718247834;
out[156] = -243.51832949342352;
out[157] = -116.65548718247834;
out[158] = 1;
out[159] = -116.65548718247834;
out[160] = 243.51832949342352;
out[161] = 116.65548718247834;
out[162] = -243.51832949342352;
out[163] = 1;
out[164] = 243.51832949342352;
out[165] = 116.65548718247834;
out[166] = -243.51832949342352;
out[167] = -116.65548718247834;
out[168] = 1;
out[169] = -116.65548718247834;
out[170] = 243.51832949342352;
out[171] = 116.65548718247834;
out[172] = -243.51832949342352;
out[173] = 1;
out[174] = 243.51832949342352;
out[175] = 116.65548718247834;
out[176] = -243.51832949342352;
out[177] = -116.65548718247834;
out[178] = 1;
out[179] = -116.65548718247834;
out[180] = 243.51832949342352;
out[181] = 116.65548718247834;
out[182] = -243.51832949342352;
out[183] = 1;
out[184] = 243.51832949342352;
out[185] = 116.65548718247834;
out[186] = -243.51832949342352;
out[187] = -116.65548718247834;
out[188] = 1;
out[189] = -116.65548718247834;
out[190] = 243.51832949342352;
out[191] = 116.65548718247834;
out[192] = -243.51832949342352;
out[193] = 1;
out[194] = 243.51832949342352;
out[195] = 116.65548718247834;
out[196] = -243.51832949342352;
out[197] = -116.65548718247834;
out[198] = 1;
out[199] = -116.65548718247834;
out[200] = 243.51832949342352;
out[201] = 116.65548718247834;
out[202] = -243.51832949342352;
out[203] = 1;
out[204] = 243.51832949342352;
out[205] = 116.65548718247834;
out[206] = -243.51832949342352;
out[207] = -116.65548718247834;
out[208] = 1;
out[209] = -116.65548718247834;
out[210] = 243.51832949342352;
out[211] = 116.65548718247834;
out[212] = -243.51832949342352;
out[213] = 1;
out[214] = 243.51832949342352;
out[215] = 116.65548718247834;
out[216] = -243.51832949342352;
out[217] = -116.65548718247834;
out[218] = 1;
out[219] = -116.65548718247834;
out[220] = 243.51832949342352;
out[221] = 116.65548718247834;
out[222] = -243.51832949342352;
out[223] = 1;
out[260] = 1;
out[261] = 1;
out[262] = 1;
out[263] = 1;
out[264] = 1;
out[265] = 1;
out[266] = 1;
out[267] = 1;
out[304] = 1;
out[305] = 1;
out[306] = 1;
out[307] = 1;
out[308] = 1;
out[309] = 1;
out[310] = 1;
out[311] = 1;
out[313] = -1;
out[314] = 1;
out[318] = -1;
out[319] = 1;
out[323] = -1;
out[324] = 1;
out[327] = -1;
out[329] = 1;
out[331] = 1;
out[332] = 1;
out[333] = 1;
out[334] = 1;
out[335] = -1;
out[337] = -1;
out[340] = -1.0;
out[341] = 1.0;
out[345] = -1.0;
out[346] = 1.0;
out[350] = -1.0;
out[351] = 1.0;
out[354] = -1.0;
out[357] = 1.0;
out[358] = 1.0;
out[359] = 1.0;
out[360] = 1.0;
out[361] = 1.0;
out[362] = -1.0;
out[365] = -1.0;
out[367] = -1;
out[368] = 1;
out[372] = -1;
out[373] = 1;
out[377] = -1;
out[378] = 1;
out[381] = -1;
out[383] = 1;
out[385] = 1;
out[386] = 1;
out[387] = 1;
out[388] = 1;
out[389] = -1;
out[391] = -1;
out[394] = -1.0;
out[395] = 1.0;
out[399] = -1.0;
out[400] = 1.0;
out[404] = -1.0;
out[405] = 1.0;
out[408] = -1.0;
out[411] = 1.0;
out[412] = 1.0;
out[413] = 1.0;
out[414] = 1.0;
out[415] = 1.0;
out[416] = -1.0;
out[419] = -1.0;
out[422] = 1;

}



/************************************************************/

static void *_cffi_types[] = {
/*  0 */ _CFFI_OP(_CFFI_OP_FUNCTION, 8), // void()(double *, double *, double *, double *, double *, double)
/*  1 */ _CFFI_OP(_CFFI_OP_POINTER, 6), // double *
/*  2 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/*  3 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/*  4 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/*  5 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/*  6 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14), // double
/*  7 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  8 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 0), // void
};

static void _cffi_d_de_jac_ini_num_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  de_jac_ini_num_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_de_jac_ini_num_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "de_jac_ini_num_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { de_jac_ini_num_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_de_jac_ini_num_eval _cffi_d_de_jac_ini_num_eval
#endif

static void _cffi_d_de_jac_ini_up_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  de_jac_ini_up_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_de_jac_ini_up_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "de_jac_ini_up_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { de_jac_ini_up_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_de_jac_ini_up_eval _cffi_d_de_jac_ini_up_eval
#endif

static void _cffi_d_de_jac_ini_xy_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  de_jac_ini_xy_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_de_jac_ini_xy_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "de_jac_ini_xy_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { de_jac_ini_xy_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_de_jac_ini_xy_eval _cffi_d_de_jac_ini_xy_eval
#endif

static void _cffi_d_de_jac_trap_num_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  de_jac_trap_num_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_de_jac_trap_num_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "de_jac_trap_num_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { de_jac_trap_num_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_de_jac_trap_num_eval _cffi_d_de_jac_trap_num_eval
#endif

static void _cffi_d_de_jac_trap_up_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  de_jac_trap_up_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_de_jac_trap_up_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "de_jac_trap_up_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { de_jac_trap_up_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_de_jac_trap_up_eval _cffi_d_de_jac_trap_up_eval
#endif

static void _cffi_d_de_jac_trap_xy_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  de_jac_trap_xy_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_de_jac_trap_xy_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "de_jac_trap_xy_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { de_jac_trap_xy_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_de_jac_trap_xy_eval _cffi_d_de_jac_trap_xy_eval
#endif

static void _cffi_d_f_ini_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  f_ini_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_f_ini_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "f_ini_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { f_ini_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_f_ini_eval _cffi_d_f_ini_eval
#endif

static void _cffi_d_f_run_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  f_run_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_f_run_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "f_run_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { f_run_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_f_run_eval _cffi_d_f_run_eval
#endif

static void _cffi_d_g_ini_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  g_ini_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_g_ini_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "g_ini_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { g_ini_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_g_ini_eval _cffi_d_g_ini_eval
#endif

static void _cffi_d_g_run_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  g_run_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_g_run_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "g_run_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { g_run_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_g_run_eval _cffi_d_g_run_eval
#endif

static void _cffi_d_h_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  h_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_h_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "h_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { h_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_h_eval _cffi_d_h_eval
#endif

static void _cffi_d_sp_jac_ini_num_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  sp_jac_ini_num_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_sp_jac_ini_num_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "sp_jac_ini_num_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { sp_jac_ini_num_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_sp_jac_ini_num_eval _cffi_d_sp_jac_ini_num_eval
#endif

static void _cffi_d_sp_jac_ini_up_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  sp_jac_ini_up_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_sp_jac_ini_up_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "sp_jac_ini_up_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { sp_jac_ini_up_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_sp_jac_ini_up_eval _cffi_d_sp_jac_ini_up_eval
#endif

static void _cffi_d_sp_jac_ini_xy_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  sp_jac_ini_xy_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_sp_jac_ini_xy_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "sp_jac_ini_xy_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { sp_jac_ini_xy_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_sp_jac_ini_xy_eval _cffi_d_sp_jac_ini_xy_eval
#endif

static void _cffi_d_sp_jac_trap_num_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  sp_jac_trap_num_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_sp_jac_trap_num_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "sp_jac_trap_num_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { sp_jac_trap_num_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_sp_jac_trap_num_eval _cffi_d_sp_jac_trap_num_eval
#endif

static void _cffi_d_sp_jac_trap_up_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  sp_jac_trap_up_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_sp_jac_trap_up_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "sp_jac_trap_up_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { sp_jac_trap_up_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_sp_jac_trap_up_eval _cffi_d_sp_jac_trap_up_eval
#endif

static void _cffi_d_sp_jac_trap_xy_eval(double * x0, double * x1, double * x2, double * x3, double * x4, double x5)
{
  sp_jac_trap_xy_eval(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_sp_jac_trap_xy_eval(PyObject *self, PyObject *args)
{
  double * x0;
  double * x1;
  double * x2;
  double * x3;
  double * x4;
  double x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "sp_jac_trap_xy_eval", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { sp_jac_trap_xy_eval(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_sp_jac_trap_xy_eval _cffi_d_sp_jac_trap_xy_eval
#endif

static const struct _cffi_global_s _cffi_globals[] = {
  { "de_jac_ini_num_eval", (void *)_cffi_f_de_jac_ini_num_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_de_jac_ini_num_eval },
  { "de_jac_ini_up_eval", (void *)_cffi_f_de_jac_ini_up_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_de_jac_ini_up_eval },
  { "de_jac_ini_xy_eval", (void *)_cffi_f_de_jac_ini_xy_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_de_jac_ini_xy_eval },
  { "de_jac_trap_num_eval", (void *)_cffi_f_de_jac_trap_num_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_de_jac_trap_num_eval },
  { "de_jac_trap_up_eval", (void *)_cffi_f_de_jac_trap_up_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_de_jac_trap_up_eval },
  { "de_jac_trap_xy_eval", (void *)_cffi_f_de_jac_trap_xy_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_de_jac_trap_xy_eval },
  { "f_ini_eval", (void *)_cffi_f_f_ini_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_f_ini_eval },
  { "f_run_eval", (void *)_cffi_f_f_run_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_f_run_eval },
  { "g_ini_eval", (void *)_cffi_f_g_ini_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_g_ini_eval },
  { "g_run_eval", (void *)_cffi_f_g_run_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_g_run_eval },
  { "h_eval", (void *)_cffi_f_h_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_h_eval },
  { "sp_jac_ini_num_eval", (void *)_cffi_f_sp_jac_ini_num_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_sp_jac_ini_num_eval },
  { "sp_jac_ini_up_eval", (void *)_cffi_f_sp_jac_ini_up_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_sp_jac_ini_up_eval },
  { "sp_jac_ini_xy_eval", (void *)_cffi_f_sp_jac_ini_xy_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_sp_jac_ini_xy_eval },
  { "sp_jac_trap_num_eval", (void *)_cffi_f_sp_jac_trap_num_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_sp_jac_trap_num_eval },
  { "sp_jac_trap_up_eval", (void *)_cffi_f_sp_jac_trap_up_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_sp_jac_trap_up_eval },
  { "sp_jac_trap_xy_eval", (void *)_cffi_f_sp_jac_trap_xy_eval, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_sp_jac_trap_xy_eval },
};

static const struct _cffi_type_context_s _cffi_type_context = {
  _cffi_types,
  _cffi_globals,
  NULL,  /* no fields */
  NULL,  /* no struct_unions */
  NULL,  /* no enums */
  NULL,  /* no typenames */
  17,  /* num_globals */
  0,  /* num_struct_unions */
  0,  /* num_enums */
  0,  /* num_typenames */
  NULL,  /* no includes */
  9,  /* num_types */
  0,  /* flags */
};

#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit_oc_4bus4wire2src_cffi(const void *p[])
{
    p[0] = (const void *)0x2601;
    p[1] = &_cffi_type_context;
#if PY_MAJOR_VERSION >= 3
    return NULL;
#endif
}
#  ifdef _MSC_VER
     PyMODINIT_FUNC
#  if PY_MAJOR_VERSION >= 3
     PyInit_oc_4bus4wire2src_cffi(void) { return NULL; }
#  else
     initoc_4bus4wire2src_cffi(void) { }
#  endif
#  endif
#elif PY_MAJOR_VERSION >= 3
PyMODINIT_FUNC
PyInit_oc_4bus4wire2src_cffi(void)
{
  return _cffi_init("oc_4bus4wire2src_cffi", 0x2601, &_cffi_type_context);
}
#else
PyMODINIT_FUNC
initoc_4bus4wire2src_cffi(void)
{
  _cffi_init("oc_4bus4wire2src_cffi", 0x2601, &_cffi_type_context);
}
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif
